/*******************************************************************************
 * Copyright (c) 2022  Xilinx, Inc. All rights reserved.
 * All rights reserved. This software is made available under
 * the terms of the Eclipse Public License 2.0
 * which accompanies this distribution, and is available at
 * https://www.eclipse.org/legal/epl-2.0/
 *
 * Xilinx, Inc.
 * Copyright 2022  Xilinx, Inc. All Rights Reserved.
 *******************************************************************************/
package org.eclipse.tcf.internal.dap;

import org.apache.commons.lang3.tuple.ImmutablePair;


import java.util.*;

import static org.apache.commons.lang3.StringUtils.isBlank;

public class TCFDebugAdapterConfiguration {

	private static final String KEY_TARGET = "target";

	private static final String KEY_HOST = "host";
	private static final String KEY_PORT = "port";
	private static final String KEY_FILE = "file";
	private static final String KEY_REMOTE_FILE_PATH = "remoteFilePath";
	private static final String KEY_CONTEXT = "context";

	private static final String KEY_PATH_MAP = "pathMap";
	private static final String KEY_SOURCE = "source";
	private static final String KEY_PATTERN = "pattern";
	private static final String KEY_DESTINATION = "destination";
	private static final String KEY_ENV = "env";
	private static final String KEY_COPY_TO_TARGET = "copyToTarget";
	private static final String KEY_COPY_FROM_TARGET_ON_TERMINATION = "copyFromTargetOnTermination";
	private static final String KEY_REPLACE_ENV = "replaceNativeEnv";
	private static final String KEY_ARGS = "arguments";
	private static final String KEY_STOP_AT_MAIN = "stopAtMain";
	private static final String KEY_NO_DEBUG = "noDebug";
	private static final String KEY_STOP_AT_PROGRAM_ENTRY = "stopAtProgramEntry";
	private static final String KEY_WORKING_DIRECTORY = "workingDirectory";
	private static final String KEY_AUTO_ATTACH_CHILDREN = "autoAttachProcessChildren";

	private final String cmdArgs;
	private final Boolean replaceEnv;
	private final Map<String, String> env;
	private final Map<String, String> copy;
	private final Map<String, String> copyOnTermination;
	private final String workingDirectory;
	private final Boolean stopAtMain;
	private final Boolean stopAtProgramEntry;
	private final Boolean autoAttachProcessChildren;

	private final String hostName;

	private final String  port;
	private final String applicationFile;
	private final String remoteFile;
	private final String context;
	private final List<ImmutablePair<String, String>> pathMapEntriesList;
	private final Boolean isNoDebug;

	public TCFDebugAdapterConfiguration(String hostName, String port, String applicationFile, String remoteFile,
										String context, String cmdArgs, Boolean replaceEnv, Map<String, String> env, Map<String, String> copyToTarget,
										Map<String, String> copyFromTargetOnTermination, String workingDir, Boolean stopAtMain, Boolean stopAtEntry,
										Boolean autoAttachProcessChildren, Boolean isNoDebug,List<ImmutablePair<String, String>> pathMap) {
		super();
		this.hostName = hostName;
		this.isNoDebug = isNoDebug;
		this.port = port;
		this.applicationFile = applicationFile;
		this.remoteFile = remoteFile;
		this.context = context;
		this.cmdArgs = cmdArgs;
		this.replaceEnv = replaceEnv;
		this.env = env;
		this.copy = copyToTarget;
		this.copyOnTermination = copyFromTargetOnTermination;
		this.workingDirectory = workingDir;
		this.stopAtMain =stopAtMain;
		this.stopAtProgramEntry = stopAtEntry;
		this.autoAttachProcessChildren = autoAttachProcessChildren;
		this.pathMapEntriesList= pathMap;
	}
	public static String getKeyTarget() {
		return KEY_TARGET;
	}
	public List<ImmutablePair<String, String>> getPathMapEntriesList() {
		return pathMapEntriesList;
	}


	public String getWorkingDirectory() {
		return workingDirectory;
	}

	public Boolean getStopAtMain() {
		return stopAtMain;
	}
	public Boolean getIsNoDebug() {
		return isNoDebug;
	}

	public Boolean getStopAtProgramEntry() {
		return stopAtProgramEntry;
	}
	public Boolean getAutoAttachProcessChildren() { return autoAttachProcessChildren; }

	public String getHostName() {
		return hostName;
	}
	public String getCmdArgs() {
		return cmdArgs;
	}

	public Boolean getReplaceEnv() {
		return replaceEnv;
	}
	public Map<String, String> getCopyOnTermination() {
		return copyOnTermination;
	}
	public Map<String, String> getEnv() {
		return env;
	}

	public String getContext() {
		return context;
	}

	public String getPort() {
		return port;
	}


	public String getApplicationFile() {
		return applicationFile;
	}


	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder();
		builder.append("TCFDebugAdapterConfiguration [hostName=").append(hostName).append(", port=").append(port)
				.append(", applicationFile=").append(applicationFile).append("]");
		return builder.toString();
	}

	@SuppressWarnings("unchecked")
	public static TCFDebugAdapterConfiguration fromArgs(Map<String, Object> args) {
		Boolean isNoDebug = Optional.ofNullable(args.get(KEY_NO_DEBUG)).map(Boolean.class::cast).orElse(false);
		String cmdArgs = Optional.ofNullable(args.get(KEY_ARGS)).map(String.class::cast).orElse("");
		String workingDir = Optional.ofNullable(args.get(KEY_WORKING_DIRECTORY)).map(String.class::cast).orElse("");
		Boolean stopAtMain = Optional.ofNullable(args.get(KEY_STOP_AT_MAIN)).map(Boolean.class::cast).orElse(true);
		Boolean stopAtEntry = Optional.ofNullable(args.get(KEY_STOP_AT_PROGRAM_ENTRY)).map(Boolean.class::cast).orElse(false);
		Boolean autoAttachChildren = Optional.ofNullable(args.get(KEY_AUTO_ATTACH_CHILDREN)).map(Boolean.class::cast).orElse(true);
		Boolean replaceEnv = Optional.ofNullable(args.get(KEY_REPLACE_ENV)).map(Boolean.class::cast).orElse(false);
		Map<String, String> env = Optional.ofNullable(args.get(KEY_ENV)).map(Map.class::cast).orElse(new HashMap<>());
		Map<String, Object> target = Optional.ofNullable(args.get(KEY_TARGET)).map(Map.class::cast).orElseThrow();
		String host = Optional.ofNullable(target.get(KEY_HOST)).map(String.class::cast).orElseThrow();
		Long port = Optional.ofNullable(target.get(KEY_PORT)).map(Double.class::cast).map(Double::longValue)
				.orElseThrow();
		String portStr = String.valueOf(port.intValue());
		String file = Optional.ofNullable(target.get(KEY_FILE)).map(String.class::cast).orElseThrow();
		String remoteFilePath = Optional.ofNullable(target.get(KEY_REMOTE_FILE_PATH)).map(String.class::cast).orElse("");
		String context = Optional.ofNullable(target.get(KEY_CONTEXT)).map(String.class::cast).orElse("");
		Object[] objects = new Object[0];
		List<Object> pathMapList = Optional.ofNullable(args.get(KEY_PATH_MAP)).map(List.class::cast).orElse(new ArrayList<>());
		List<ImmutablePair<String,String>> pathMapEntries = new ArrayList<>();
		for( Object pathMapEntry : pathMapList ){
			Map<String,String> oneEntry = Optional.ofNullable(pathMapEntry).map(Map.class::cast).orElse(new HashMap<>());
			String source = Optional.ofNullable(oneEntry.get(KEY_SOURCE)).map(String.class::cast).orElseThrow();
			String dest = Optional.ofNullable(oneEntry.get(KEY_DESTINATION)).map(String.class::cast).orElseThrow();
			pathMapEntries.add(ImmutablePair.of(source,dest));
		}

		List<Object> copyToTargetList = Optional.ofNullable(args.get(KEY_COPY_TO_TARGET)).map(List.class::cast).orElse(new ArrayList<>());
		Map<String, String> copyToTarget = new HashMap<>();
		for( Object copyToTargetEntry : copyToTargetList ){
			Map<String,String> oneEntry = Optional.ofNullable(copyToTargetEntry).map(Map.class::cast).orElse(new HashMap<>());
			String source = Optional.ofNullable(oneEntry.get(KEY_SOURCE)).map(String.class::cast).orElseThrow();
			String dest = Optional.ofNullable(oneEntry.get(KEY_DESTINATION)).map(String.class::cast).orElseThrow();
			copyToTarget.put(source,dest);
		}
		List<Object> copyFromTargetList = Optional.ofNullable(args.get(KEY_COPY_FROM_TARGET_ON_TERMINATION)).map(List.class::cast).orElse(new ArrayList<>());
		Map<String, String> copyFromTarget = new HashMap<>();
		for( Object copyToTargetEntry : copyFromTargetList ){
			Map<String,String> oneEntry = Optional.ofNullable(copyToTargetEntry).map(Map.class::cast).orElse(new HashMap<>());
			String source = Optional.ofNullable(oneEntry.get(KEY_SOURCE)).map(String.class::cast).orElse("");
			String pattern = Optional.ofNullable(oneEntry.get(KEY_PATTERN)).map(String.class::cast).orElse("");
			String dest = Optional.ofNullable(oneEntry.get(KEY_DESTINATION)).map(String.class::cast).orElseThrow();
			if(isBlank(source) && isBlank(pattern)) {
				throw new RuntimeException("Either source or pattern must be non empty in " + KEY_COPY_FROM_TARGET_ON_TERMINATION );
			}
			copyFromTarget.put((source.isEmpty() ? pattern: source),dest);
		}
		if(isNoDebug){
			stopAtMain = false;
			stopAtEntry = false;
			autoAttachChildren = false;
		}


		return new TCFDebugAdapterConfiguration(host, portStr, file,remoteFilePath,
				context,cmdArgs,replaceEnv,env,copyToTarget,copyFromTarget,workingDir,
				stopAtMain,stopAtEntry,autoAttachChildren,isNoDebug,pathMapEntries);
	}

	public String getRemoteFile() {
		return remoteFile;
	}

	public Map<String, String> getCopy() {
		return copy;
	}
}
