/*******************************************************************************
 * Copyright (c) 2007-2020 Wind River Systems, Inc. and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License 2.0
 * which accompanies this distribution, and is available at
 * https://www.eclipse.org/legal/epl-2.0/
 *
 * Contributors:
 *     Wind River Systems - initial API and implementation
 *******************************************************************************/

/*******************************************************************************
 * Modified by Xilinx,Inc
 * Copyright 2022 Xilinx, Inc. All Rights Reserved. Modifications
 * are made available under the terms of the Eclipse Public License 2.0
 * which accompanies this distribution, and is available at
 * https://www.eclipse.org/legal/epl-2.0/
 *
 *******************************************************************************/

package org.eclipse.tcf.internal.dap;

import org.apache.commons.lang3.tuple.ImmutablePair;
import org.eclipse.lsp4j.debug.*;
import org.eclipse.lsp4j.debug.services.IDebugProtocolClient;
import org.eclipse.lsp4j.jsonrpc.RemoteEndpoint;
import org.eclipse.tcf.core.ILaunch;
import org.eclipse.tcf.core.UserDefPeer;
import org.eclipse.tcf.internal.core.ServiceManager;
import org.eclipse.tcf.internal.dap.actions.TCFAction;
import org.eclipse.tcf.internal.dap.breakpoint.TCFBreakpointHandler;
import org.eclipse.tcf.protocol.*;
import org.eclipse.tcf.services.*;
import org.eclipse.tcf.util.TCFDataCache;
import org.eclipse.tcf.util.TCFMemoryRegion;
import org.eclipse.tcf.util.TCFPathMapRule;

import java.io.*;
import java.nio.file.FileVisitOption;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.*;
import java.util.concurrent.CompletableFuture;
import java.util.logging.Logger;

import static java.nio.file.FileVisitOption.FOLLOW_LINKS;
import static org.apache.commons.lang3.StringUtils.isBlank;
import static org.eclipse.tcf.services.IFileSystem.*;

public class TCFLaunch implements ILaunch {


    public TCFDebugAdapterConfiguration getDebugAdapterConfiguration() {
        return debugAdapterConfiguration;
    }

    private final IDebugProtocolClient client;
    private final RemoteEndpoint remoteEndpoint;
    private TCFDebugAdapterConfiguration debugAdapterConfiguration;
    protected static final Logger LOGGER = Logger.getLogger(TCFLaunch.class.getSimpleName());
    private final HashMap<String,TCFDataCache<String[]>> context_query_cache = new HashMap<String,TCFDataCache<String[]>>();
    private final HashMap<String,TCFAction> active_actions = new HashMap<String,TCFAction>();
    private final HashMap<String,LinkedList<TCFAction>> context_action_queue = new HashMap<String,LinkedList<TCFAction>>();
    private final HashMap<String,Long> context_action_timestamps = new HashMap<String,Long>();
    private final HashSet<Object> pending_clients = new HashSet<Object>();
    private long pending_clients_timestamp;
    private final Collection<ActionsListener> action_listeners = new ArrayList<ActionsListener>();
    private TCFDataCache<IPathMap.PathMapRule[]> target_path_map;
    private final LinkedList<LaunchStep> launch_steps = new LinkedList<LaunchStep>();
    private List<IPathMap.PathMapRule> host_path_map = new ArrayList<IPathMap.PathMapRule>();
    private final HashMap<String,String> process_stream_ids = new HashMap<String,String>();
    private final HashMap<String, IProcesses.ProcessContext> attached_processes = new HashMap<String, IProcesses.ProcessContext>();
    private final HashMap<String,String> uart_tx_stream_ids = new HashMap<String,String>();
    private final HashMap<String,String> uart_rx_stream_ids = new HashMap<String,String>();
    private final HashSet<String> disconnected_stream_ids = new HashSet<String>();
    private static final Collection<LaunchListener> listeners = new ArrayList<LaunchListener>();
    private static LaunchListener[] listeners_array;
    private InputStream stdin_file_stream;
    private OutputStream stdout_file_stream;
    private String dprintf_stream_id;
    private boolean connecting;
    private boolean disconnecting;
    private boolean disconnected;
    private boolean shutdown;
    private boolean last_context_exited;
    private Runnable update_memory_maps;
    private final HashMap<String,String> process_env = new HashMap<String,String>();
    private boolean can_terminate_attached = false;
    private IProcesses.ProcessContext process;
    private boolean process_exited;
    private int process_exit_code;
    private Set<String> context_filter;
    // private HashMap<String,IStorage> target_path_mapping_cache = new HashMap<String,IStorage>();
    private Collection<Map<String,Object>> process_signals;
    private IChannel channel;
    private Throwable error;
    private long actions_interval;
    private TCFBreakpointsStatus breakpoints_status;
    private TCFBreakpointStatusListener bp_status_listener;
    private TCFBreakpointHandler breakpointHandler;

    public TCFLaunch(TCFDebugAdapterConfiguration debugAdapterConfiguration, IDebugProtocolClient client, RemoteEndpoint remoteEndpoint) {
        this.debugAdapterConfiguration = debugAdapterConfiguration;
        this.client = client;
        this.remoteEndpoint = remoteEndpoint;
        for (LaunchListener l : getListeners()) l.onCreated(TCFLaunch.this);
    }

    /**
     * Register a launch listener.
     * @param listener - client object implementing TCFLaunch.LaunchListener interface.
     */
    public static void addListener(LaunchListener listener) {
        assert Protocol.isDispatchThread();
        listeners.add(listener);
        listeners_array = null;
    }

    public IDebugProtocolClient getClient() {
        return client;
    }
    public RemoteEndpoint getRemoteEndpoint(){
        return remoteEndpoint;

    }

    /**
     * Remove a launch listener.
     * @param listener - client object implementing TCFLaunch.LaunchListener interface.
     */
    public static void removeListener(LaunchListener listener) {
        assert Protocol.isDispatchThread();
        listeners.remove(listener);
        listeners_array = null;
    }
    private final IRunControl.RunControlListener rc_listener = new IRunControl.RunControlListener() {

        private void flushContextQueryCache() {
            for (TCFDataCache<?> c : context_query_cache.values()) c.reset();
        }

        public void contextAdded(IRunControl.RunControlContext[] contexts) {
            flushContextQueryCache();
        }

        public void contextChanged(IRunControl.RunControlContext[] contexts) {
            flushContextQueryCache();
        }

        public void contextRemoved(String[] context_ids) {
            flushContextQueryCache();
        }

        public void contextSuspended(String context, String pc, String reason, Map<String, Object> params) {
        }

        public void contextResumed(String context) {
        }

        public void containerSuspended(String context, String pc, String reason, Map<String, Object> params, String[] suspended_ids) {
        }

        public void containerResumed(String[] context_ids) {
        }

        public void contextException(String context, String msg) {
        }
    };

    /**
     * Get TCF communication channel that is used by the launch.
     * Thread safe method.
     */
    public IChannel getChannel() {
        return channel;
    }

    /**
     * If the launch has started a remote process, return the process information.
     * Starting a process is optional and not applicable to all launches.
     * @return remote process information or null.
     */
    public IProcesses.ProcessContext getProcessContext() {
        return process;
    }

    /**
     * Add an object to the set of pending clients.
     * Actions execution will be delayed until the set is empty,
     * but not longer then 1 second.
     * @param client
     */
    public void addPendingClient(Object client) {
        synchronized (pending_clients) {
            pending_clients.add(client);
            pending_clients_timestamp = System.currentTimeMillis();
        }
    }

    /**
     * Remove an object from the set of pending clients.
     * Actions execution resumes when the set becomes empty.
     * @param client
     */
    public void removePendingClient(Object client) {
        synchronized (pending_clients) {
            if (pending_clients.remove(client) && pending_clients.size() == 0) {
                pending_clients_timestamp = System.currentTimeMillis();
            }
        }
    }


    private abstract class LaunchStep implements Runnable {

        LaunchStep() {
            launch_steps.add(this);
        }

        abstract void start() throws Exception;

        void done() {
            if (channel.getState() != IChannel.STATE_OPEN) return;
            try {
                launch_steps.removeFirst().start();
            }
            catch (Throwable x) {
                channel.terminate(x);
            }
        }

        public void run() {
            done();
        }
    }
    public static class PathMapRule extends TCFPathMapRule {

        public PathMapRule(Map<String,Object> props) {
            super(props);
        }

        /* (non-Javadoc)
         * @see java.lang.Object#equals(java.lang.Object)
         */
        @Override
        public boolean equals(Object obj) {
            boolean equal = super.equals(obj);
            if (!equal && obj instanceof PathMapRule) {
                return this.toString().equals(((PathMapRule)obj).toString());
            }
            return equal;
        }

        @Override
        public String toString() {
            StringBuffer bf = new StringBuffer();
            Map<String,Object> props = getProperties();
            String[] keySet = props.keySet().toArray(new String[props.size()]);
            Arrays.sort(keySet);
            for (String nm : keySet) {
                Object o = props.get(nm);
                if (o != null) {
                    bf.append(nm);
                    bf.append('=');
                    String s = o.toString();
                    for (int i = 0; i < s.length(); i++) {
                        char ch = s.charAt(i);
                        if (ch >= ' ' && ch != '|' && ch != '\\') {
                            bf.append(ch);
                        }
                        else {
                            bf.append('\\');
                            bf.append((int)ch);
                            bf.append(';');
                        }
                    }
                    bf.append('|');
                }
            }
            bf.append('|');
            return bf.toString();
        }
    }
    private final IStreams.StreamsListener streams_listener = new IStreams.StreamsListener() {

        public void created(String stream_type, String stream_id, String context_id) {
            disconnected_stream_ids.remove(stream_id);
            if (stream_type.equals("UART-TX")) {
                uart_tx_stream_ids.put(stream_id, context_id);
                readStream(context_id, stream_id, 0);
            }
            else if (stream_type.equals("UART-RX")) {
                uart_rx_stream_ids.put(stream_id, context_id);
            }
            else {
                process_stream_ids.put(stream_id, context_id);
                disconnectUnusedStreams();
            }
        }

        public void disposed(String stream_type, String stream_id) {
            disconnected_stream_ids.add(stream_id);
        }
    };
    private static LaunchListener[] getListeners() {
        if (listeners_array != null) return listeners_array;
        return listeners_array = listeners.toArray(new LaunchListener[listeners.size()]);
    }


    private void readStream(final String ctx_id, final String id, final int no) {
        if (ctx_id != null) {
            // Force creation of console
            for (LaunchListener l : getListeners()) l.onProcessOutput(this, ctx_id, no, null);
            // Read console input file
            if (stdin_file_stream != null) {
                try {
                    byte[] buf = new byte[256];
                    for (;;) {
                        int rd = stdin_file_stream.read(buf);
                        if (rd <= 0) break;
                        writeProcessInputStream(ctx_id, buf, 0, rd);
                    }
                    stdin_file_stream.close();
                }
                catch (Exception x) {
                    // Activator.log("Cannot read console capture file", x);
                    LOGGER.info("Cannot read console capture file" +  x.toString());
                }
                stdin_file_stream = null;
            }
        }
        final IStreams streams = getService(IStreams.class);
        IStreams.DoneRead done = new IStreams.DoneRead() {
            public void doneRead(IToken token, Exception error, int lost_size, byte[] data, boolean eos) {
                if (lost_size > 0) {
                    Exception x = new IOException("Process output data lost due buffer overflow");
                    for (LaunchListener l : getListeners()) l.onProcessStreamError(TCFLaunch.this, ctx_id, no, x, lost_size);
                }
                if (data != null && data.length > 0) {
                    for (LaunchListener l : getListeners()) l.onProcessOutput(TCFLaunch.this, ctx_id, no, data);
                    if (stdout_file_stream != null && data != null) {
                        try {
                            stdout_file_stream.write(data);
                        }
                        catch (IOException x) {
                           //  Activator.log("Cannot write console capture file", x);
                            LOGGER.info("Cannot write console capture file"+  x.toString());
                            try {
                                stdout_file_stream.close();
                            }
                            catch (IOException y) {
                                LOGGER.info("Cannot close console capture file" + y.toString());
                                // Activator.log("Cannot close console capture file", y);
                            }
                            stdout_file_stream = null;
                        }
                    }
                }
                if (disconnected_stream_ids.contains(id)) return;
                if (error != null) {
                    for (LaunchListener l : getListeners()) l.onProcessStreamError(TCFLaunch.this, ctx_id, no, error, 0);
                    disconnected_stream_ids.add(id);
                }
                if (!eos && error == null) {
                    streams.read(id, 0x1000, this);
                }
            }
        };
        streams.read(id, 0x1000, done);
        streams.read(id, 0x1000, done);
        streams.read(id, 0x1000, done);
        streams.read(id, 0x1000, done);
    }

    private void readProcessStreams(IProcesses.ProcessContext ctx) {
        assert attached_processes.get(ctx.getID()) == ctx;
        IStreams streams = getService(IStreams.class);
        if (streams == null) return;
        String out_id = (String)ctx.getProperties().get(IProcesses.PROP_STDOUT_ID);
        String err_id = (String)ctx.getProperties().get(IProcesses.PROP_STDERR_ID);
        if (process_stream_ids.get(out_id) != null) readStream(ctx.getID(), out_id, 0);
        if (process_stream_ids.get(err_id) != null) readStream(ctx.getID(), err_id, 0);
    }

    /**
     * Write to stdin stream of remote process that was started by the launch.
     * @param prs_id - TCF ID of the process.
     * @param buf - data to write
     * @param pos - starting position in 'buf'
     * @param len - number of bytes to write.
     * @throws Exception
     */
    public void writeProcessInputStream(final String prs_id, byte[] buf, int pos, final int len) throws Exception {
        assert Protocol.isDispatchThread();
        if (channel.getState() != IChannel.STATE_OPEN) throw new IOException("Connection closed");
        IStreams streams = getService(IStreams.class);
        if (streams == null) throw new IOException("Streams service not available");
        IProcesses.ProcessContext ctx = attached_processes.get(prs_id);
        if (ctx != null) {
            final String id = (String)ctx.getProperties().get(IProcesses.PROP_STDIN_ID);
            if (process_stream_ids.get(id) == null) throw new IOException("Input stream not available");
            streams.write(id, buf, pos, len, new IStreams.DoneWrite() {
                public void doneWrite(IToken token, Exception error) {
                    if (error == null) return;
                    if (process_stream_ids.get(id) == null) return;
                    for (LaunchListener l : getListeners()) l.onProcessStreamError(TCFLaunch.this, prs_id, 0, error, len);
                    disconnectStream(id);
                }
            });
            return;
        }
        for (final String rx_id : uart_rx_stream_ids.keySet()) {
            if (!prs_id.equals(uart_rx_stream_ids.get(rx_id))) continue;
            streams.write(rx_id, buf, pos, len, new IStreams.DoneWrite() {
                public void doneWrite(IToken token, Exception error) {
                    if (error == null) return;
                    if (uart_rx_stream_ids.get(rx_id) == null) return;
                    for (LaunchListener l : getListeners()) l.onProcessStreamError(TCFLaunch.this, prs_id, 0, error, len);
                    disconnectStream(rx_id);
                }
            });
            return;
        }
        throw new IOException("No target process");
    }

    /**
     * Terminate the launch because of fatal error.
     * @param x - the error object.
     */
    public void setError(Throwable x) {
        error = x;
        if (x != null) {
            if (channel != null && channel.getState() == IChannel.STATE_OPEN) {
                channel.terminate(x);
            }
            else if (!connecting) {
                disconnected = true;
                shutdown = true;
            }
        }
        // fireChanged();
    }

    private void disconnectUnusedStreams() {
        if (connecting) return;
        HashSet<String> set = new HashSet<String>();
        for (IProcesses.ProcessContext ctx : attached_processes.values()) {
            set.add((String)ctx.getProperties().get(IProcesses.PROP_STDIN_ID));
            set.add((String)ctx.getProperties().get(IProcesses.PROP_STDOUT_ID));
            set.add((String)ctx.getProperties().get(IProcesses.PROP_STDERR_ID));
        }
        for (String id : process_stream_ids.keySet().toArray(new String[process_stream_ids.size()])) {
            if (!set.contains(id)) disconnectStream(id);
        }
    }

    private void disconnectStream(String id) {
        assert process_stream_ids.get(id) != null;
        process_stream_ids.remove(id);
        if (channel.getState() != IChannel.STATE_OPEN) return;
        disconnected_stream_ids.add(id);
        IStreams streams = getService(IStreams.class);
        streams.disconnect(id, new IStreams.DoneDisconnect() {
            public void doneDisconnect(IToken token, Exception error) {
                if (channel.getState() != IChannel.STATE_OPEN) return;
                if (error != null) channel.terminate(error);
            }
        });
    }

    /**
     * Launch object handles queue of user actions, like debugger stepping commands.
     * ActionsListener allows clients to be notified when an action execution is started and finished.
     */
    public interface ActionsListener {

        public void onContextActionStart(TCFAction action);

        public void onContextActionResult(String id, String result);

        public void onContextActionDone(TCFAction action);
    }

    /**
     * Clients can use LaunchListener interface for notifications of launch being created, connected or
     * disconnected. The interface also allows to receive remote process output.
     */
    public interface LaunchListener {

        public void onCreated(TCFLaunch launch);

        public void onConnected(TCFLaunch launch);

        public void onDisconnected(TCFLaunch launch);

        public void onProcessOutput(TCFLaunch launch, String process_id, int stream_id, byte[] data);

        public void onProcessStreamError(
                TCFLaunch launch, String process_id, int stream_id,
                Exception error, int lost_size);
    }

    private void onDisconnected(Throwable error) {
        // The method is called when TCF channel is closed.
        assert !disconnected;
        assert !shutdown;
        this.error = error;
        // breakpoints_status = null;
        connecting = false;
        disconnected = true;
        for (LaunchListener l : getListeners()) l.onDisconnected(this);
        for (TCFDataCache<?> c : context_query_cache.values()) c.dispose();
        context_query_cache.clear();


        // closeConsoleCaptureFiles();
//        if (DebugPlugin.getDefault() != null) fireChanged();
//        if (launch_task != null) launch_task.done(false);
//        launch_monitor = null;
//        launch_task = null;
//        runShutdownSequence(new Runnable() {
//            public void run() {
//                shutdown = true;
//                fireTerminate();
//                for (TCFTask<Boolean> tsk : disconnect_wait_list) {
//                    tsk.done(Boolean.TRUE);
//                }
//                disconnect_wait_list.clear();
//            }
//        });
//        // Log severe exceptions: bug 386067
//        if (error instanceof RuntimeException) {
//            Activator.log("Channel disconnected with error", error);
//        }
    }


    /**
     * Add a context action to actions queue.
     * Examples of context actions are resume/suspend/step commands,
     * which were requested by a user.
     * @param action
     */
    public void addContextAction(TCFAction action) {
        assert Protocol.isDispatchThread();
        String id = action.getContextID();
        LinkedList<TCFAction> list = context_action_queue.get(id);
        if (list == null) context_action_queue.put(id, list = new LinkedList<TCFAction>());
        int priority = action.getPriority();
        for (ListIterator<TCFAction> i = list.listIterator();;) {
            if (i.hasNext()) {
                if (priority <= i.next().getPriority()) continue;
                i.previous();
            }
            i.add(action);
            break;
        }
        startAction(id);
    }
    private long getActionTimeStamp(String id) {
        Long l = context_action_timestamps.get(id);
        if (l == null) return 0;
        return l.longValue();
    }
    /**
     * Set action result for given context ID.
     * Action results are usually presented to a user same way as context suspend reasons.
     * @param id - debug context ID.
     * @param result - a string to be shown to user.
     */
    public void setContextActionResult(String id, String result) {
        assert Protocol.isDispatchThread();
        for (ActionsListener l : action_listeners) l.onContextActionResult(id, result);
    }


    /**
     * Remove an action from the queue.
     * The method should be called when the action execution is done.
     * @param action
     */
    public void removeContextAction(TCFAction action) {
        assert Protocol.isDispatchThread();
        String id = action.getContextID();
        assert active_actions.get(id) == action;
        active_actions.remove(id);
        for (ActionsListener l : action_listeners) l.onContextActionDone(action);
        startAction(id);
    }

    /**
     * Remove all actions from the queue of a debug context.
     * @param id - debug context ID.
     */
    public void removeContextActions(String id) {
        assert Protocol.isDispatchThread();
        context_action_queue.remove(id);
        context_action_timestamps.remove(id);
    }

    /**
     * Get action queue size of a debug context.
     * @param id - debug context ID.
     * @return count of pending actions.
     */
    public int getContextActionsCount(String id) {
        assert Protocol.isDispatchThread();
        LinkedList<TCFAction> list = context_action_queue.get(id);
        int n = list == null ? 0 : list.size();
        if (active_actions.get(id) != null) n++;
        return n;
    }


    private void startAction(final String id) {
        if (active_actions.get(id) != null) return;
        LinkedList<TCFAction> list = context_action_queue.get(id);
        if (list == null || list.size() == 0) return;
        final TCFAction action = list.removeFirst();
        if (list.size() == 0) context_action_queue.remove(id);
        active_actions.put(id, action);
        final long timestamp = getActionTimeStamp(id);
        long time = System.currentTimeMillis();
        Protocol.invokeLater(timestamp + actions_interval - time, new Runnable() {
            public void run() {
                if (active_actions.get(id) != action) return;
                long time = System.currentTimeMillis();
                synchronized (pending_clients) {
                    if (pending_clients.size() > 0) {
                        if (time - timestamp < actions_interval + 1000) {
                            Protocol.invokeLater(20, this);
                            return;
                        }
                        pending_clients.clear();
                    }
                    else if (time < pending_clients_timestamp + 10) {
                        Protocol.invokeLater(pending_clients_timestamp + 10 - time, this);
                        return;
                    }
                }
                context_action_timestamps.put(id, time);
                for (ActionsListener l : action_listeners) l.onContextActionStart(action);
                action.run();
            }
        });
    }
    /**
     * Set minimum interval between context actions execution.
     * @param interval - minimum interval in milliseconds.
     */
    public void setContextActionsInterval(long interval) {
        actions_interval = interval;
    }

    public void doLaunch(CompletableFuture<Void> future) {
        assert Protocol.isDispatchThread();
        Map<String, String> peerAttrs = createAttributes();
        IPeer peer = new UserDefPeer(peerAttrs);
        System.out.println(peer.getAttributes());
        LOGGER.finer("TCF Peer Attributes " + peer.getAttributes());
        openChannel(peer, future);
    }
    public <V extends IService> V getService(Class<V> cls) {
        assert Protocol.isDispatchThread();
        return channel.getRemoteService(cls);
    }
    /**
     * Apply the path map to this launch channel.
     * @param done The done to invoke.
     */
    protected void applyPathMap(final Runnable done) {
        IPathMap path_map_service = getService(IPathMap.class);
        if (path_map_service == null) {
            if (done != null) done.run();
            return;
        }
        path_map_service.set(host_path_map.toArray(new IPathMap.PathMapRule[host_path_map.size()]), new IPathMap.DoneSet() {
            @Override
            public void doneSet(IToken token, Exception error) {
                if (error != null) channel.terminate(error);
                else if (done != null) done.run();
            }
        });
    }

    private  void openChannel(IPeer peer, CompletableFuture<Void> future) {
        channel = peer.openChannel();
        IChannel.IChannelListener listener = new IChannel.IChannelListener() {
            public void onChannelOpened() {
                try {
                    onConnected(future);
                } catch (Throwable x) {
                    TCFLaunch.this.channel.terminate(x);
                }
            }

            public void congestionLevel(int level) {
            }

            public void onChannelClosed(Throwable error) {
                TCFLaunch.this.channel.removeChannelListener(this);
                onDisconnected(error);
                if (error != null) {
                    LOGGER.finer("TCF channel closed with error " + error.toString());
                    error.printStackTrace();
                    future.completeExceptionally(error);
                    Protocol.invokeLater(200, new Runnable() {
                                @Override
                                public void run() {
                                    TerminatedEventArguments terminatedArgs = new TerminatedEventArguments();
                                    client.terminated(terminatedArgs);
                                }
                            }
                    );
                }
            }
        };
        channel.addChannelListener(listener);
        assert channel.getState() == IChannel.STATE_OPENING;
        connecting = true;
    }
        private  void onConnected(CompletableFuture<Void> future) {
            LOGGER.fine("TCF channel opened");

            boolean use_context_filter = !isBlank(debugAdapterConfiguration.getContext());
            if (use_context_filter) context_filter = new HashSet<String>();

            final IRunControl rc_service = getService(IRunControl.class);
            if (rc_service != null) rc_service.addListener(rc_listener);

            final IPathMap path_map_service = getService(IPathMap.class);
            if (path_map_service != null) {
                target_path_map = new TCFDataCache<IPathMap.PathMapRule[]>(channel) {
                    @Override
                    protected boolean startDataRetrieval() {
                        command = path_map_service.get(new IPathMap.DoneGet() {
                            public void doneGet(IToken token, Exception error, IPathMap.PathMapRule[] map) {
                                set(token, error, map);
                            }
                        });
                        return false;
                    }
                };
                path_map_service.addListener(new IPathMap.PathMapListener() {
                    public void changed() {
                        target_path_map.reset();
                        // target_path_mapping_cache = new HashMap<String,IStorage>();
                    }
                });
            }
            if (path_map_service != null) {
                new LaunchStep() {
                    @Override
                    void start() throws Exception {
                        readPathMapConfiguration(debugAdapterConfiguration);
                        applyPathMap(this);
                    }
                };
            }

            final IStreams streams = getService(IStreams.class);
            if (streams != null) {
                // Subscribe Streams service:
                new LaunchStep() {
                    @Override
                    void start() {
                        final Set<IToken> cmds = new HashSet<IToken>();
                        String[] nms = { IProcesses.NAME, IProcessesV1.NAME, "UART-RX", "UART-TX" };
                        for (String s : nms) {
                            cmds.add(streams.subscribe(s, streams_listener, new IStreams.DoneSubscribe() {
                                public void doneSubscribe(IToken token, Exception error) {
                                    cmds.remove(token);
                                    if (error != null) channel.terminate(error);
                                    if (cmds.size() == 0) done();
                                }
                            }));
                        }
                        if (cmds.size() == 0) done();
                    }
                };
            }
            if(!isBlank(debugAdapterConfiguration.getContext())){
                context_filter.add(debugAdapterConfiguration.getContext());
            }
            final IMemoryMap mem_map = channel.getRemoteService(IMemoryMap.class);
            if (mem_map != null) {
                // Send manual memory map items:
                new LaunchStep() {
                    @Override
                    void start() throws Exception {
                        downloadMemoryMaps(this);
                    }
                };
            }

            // Send breakpoints:
            new LaunchStep() {
                @Override
                void start() throws Exception {
                    breakpoints_status = new TCFBreakpointsStatus(TCFLaunch.this);
                    TCFBreakpointsModel.getBreakpointsModel().initialize(channel, this);
                    bp_status_listener = new TCFBreakpointStatusListener();
                    breakpointHandler = new TCFBreakpointHandler(channel);
                    bp_status_listener.setBreakpointHandler(breakpointHandler);
                }
            };

            final IDPrintf dprintf = getService(IDPrintf.class);
            if (dprintf != null) {
                // Open dprintf stream:
                new LaunchStep() {
                    @Override
                    void start() throws Exception {
                        dprintf.open(null, new IDPrintf.DoneCommandOpen() {
                            @Override
                            public void doneCommandOpen(IToken token, Exception error, final String id) {
                                if (error != null) {
                                    channel.terminate(error);
                                    return;
                                }
                                dprintf_stream_id = id;
                                streams.connect(id, new IStreams.DoneConnect() {
                                    @Override
                                    public void doneConnect(IToken token, Exception error) {
                                        if (error != null) {
                                            channel.terminate(error);
                                            return;
                                        }
                                        readStream(null, id, 0);
                                        done();
                                    }
                                });
                            }
                        });
                    }
                };
            }
            // Call client launch sequence:
            new LaunchStep() {
                @Override
                void start() {
                    runLaunchSequence(this);
                }
            };

            try {
                startRemoteProcess();
            } catch (Exception e) {
                LOGGER.info("Failed to start the remote process. " + e.toString());
                clientErr("Failed to start the remote process.");
                e.printStackTrace();
            }
            // Final launch step.
            // Notify clients:
            new LaunchStep() {
                @Override
                void start() {
                    connecting = false;
                    disconnectUnusedStreams();
                    // openConsoleCaptureFiles();
                    for (LaunchListener l : getListeners()) l.onConnected(TCFLaunch.this);
                    future.complete(null);
                        /*
                         fireChanged();
                         if (launch_task != null) launch_task.done(true);
                         launch_monitor = null;
                        launch_task = null;
                         */
                }
            };
            try {
                launch_steps.removeFirst().start();
            } catch (Exception e) {
                LOGGER.info("Failed to launch the remote process. " + e.toString());
                clientErr("Failed to launch the remote process.");
                e.printStackTrace();
            }

        }


    protected void runLaunchSequence(Runnable done) {
        done.run();
    }
    // example values: /tmp/dir1 /mnt
    // the remoteDirPath do not include the leaf directory name.
    private void addDirectoryCopyStep(String localDir,String remoteDirPath){

        EnumSet<FileVisitOption> opts = EnumSet.of(FOLLOW_LINKS);
        TCFLocalFileVisitor visitor = new TCFLocalFileVisitor();
        try {
            Files.walkFileTree(Paths.get(localDir), opts,Integer.MAX_VALUE, visitor);
        } catch (IOException e) {
            e.printStackTrace();
        }
        Path localDirPath = Paths.get(localDir);
        int localDirNameCount = localDirPath.getNameCount();
        //2  /tmp/dir1 /mnt
        //4  -> /tmp/dir1/dir2/file1  -> /mnt/dir1/dir2/file1
        //
        Map<String,Boolean> dirMap = new HashMap<>();
        for(Path localFilePath : visitor.getFilePaths()){
            int fileNameCount = localFilePath.getNameCount();
            int segmentsToBeAddedToTarget = fileNameCount - localDirNameCount + 1;
            String remoteFilePath = Paths.get(remoteDirPath,localFilePath.subpath(fileNameCount-segmentsToBeAddedToTarget,fileNameCount).toString()).toString();
            for(int i = 1; i< segmentsToBeAddedToTarget ; i++){
                int startIndex = fileNameCount-segmentsToBeAddedToTarget;
                String targetDir = Paths.get(remoteDirPath,localFilePath.subpath(startIndex,startIndex+i).toString()).toString();
                if(!dirMap.containsKey(targetDir)) {
                    addMkDirStep(targetDir);
                    dirMap.put(targetDir,true);
                }
            }
            addFileCopyStep(localFilePath.toAbsolutePath().toString(),remoteFilePath);
        }
    }
    private void addMkDirStep(String remoteDirPath){
        new LaunchStep() {
            @Override
            void start() throws Exception {
                mkDirOnRemoteTarget(remoteDirPath,this);
            }
        };
    }


    private void addFileCopyStep(String localFile,String remoteFilePath){
        new LaunchStep() {
            @Override
            void start() throws Exception {
                copyFileToRemoteTarget(localFile, remoteFilePath, true,this);
            }
        };
    }
    private void startRemoteProcess() throws Exception {
        if (!isBlank(debugAdapterConfiguration.getRemoteFile())) {
            new LaunchStep() {
                @Override
                void start() throws Exception {
                    copyFileToRemoteTarget(debugAdapterConfiguration.getApplicationFile(), debugAdapterConfiguration.getRemoteFile(), false,this);
                }
            };
            Map<String, String> filesToCopy = debugAdapterConfiguration.getCopy();
            for(Map.Entry<String,String> fileEntry : filesToCopy.entrySet()){
                File localFile = new File(fileEntry.getKey());
                Path leafName = Paths.get(fileEntry.getKey()).getFileName();
                if (localFile.isDirectory()){
                    addDirectoryCopyStep(localFile.getAbsolutePath().toString(),fileEntry.getValue());
                }else {
                    Path remoteFilePath = Paths.get(fileEntry.getValue(), leafName.toString());
                    addFileCopyStep(localFile.getAbsolutePath().toString(),remoteFilePath.toString());
                }
            }
//            Map<String, String> filesToCopyFrom = debugAdapterConfiguration.getCopyOnTermination();
//            for(Map.Entry<String,String> fileEntry : filesToCopyFrom.entrySet()) {
//                Path leafName = Paths.get(fileEntry.getKey()).getFileName();
//                String localFile = Paths.get(fileEntry.getValue(),leafName.toString()).toString();
//                new LaunchStep() {
//                    @Override
//                    void start() {
//                        copyFileFromRemoteTarget(fileEntry.getKey(), localFile, this);
//                    }
//                };
//            }

        }
        final IProcesses ps = channel.getRemoteService(IProcesses.class);
        if (ps == null) throw new Exception("Target does not provide Processes service");
        final boolean append = !debugAdapterConfiguration.getReplaceEnv();
        /*
        // Need to support append from the launch json entries.
        debugAdapterConfiguration.shouldAppend()
        debugAdapterConfiguration.getEnv();
        */


        if (append) {
            // Get system environment variables
            new LaunchStep() {
                @Override
                void start() throws Exception {
                    ps.getEnvironment(new IProcesses.DoneGetEnvironment() {
                        public void doneGetEnvironment(IToken token, Exception error, Map<String, String> env) {
                            if (error != null) {
                                channel.terminate(error);
                            } else {
                                if (env != null) process_env.putAll(env);
                                done();
                            }
                        }
                    });
                }
            };
        }
        final String dir = debugAdapterConfiguration.getWorkingDirectory(); // Working directory.
        final String args = debugAdapterConfiguration.getCmdArgs(); // command line arguments.
        final Map<String, String> env = debugAdapterConfiguration.getEnv(); // env var settings.
        // NoDebug -> true means Run.
        // By default NoDebug is false, means it respects the stop_entry, stop_at_main, attach_children settings.
        // If Nodebug is true, then stop_entry, stop_at_main, attach_children settings will have false as value.

        final boolean attach_children =  debugAdapterConfiguration.getAutoAttachProcessChildren();
        final boolean stop_at_entry = debugAdapterConfiguration.getStopAtProgramEntry(); // should stop at entry, for tcf launch in eclipse default is true.
        final boolean stop_at_main =  debugAdapterConfiguration.getStopAtMain();
        final boolean use_terminal = true;
        // Stop and pass signals.
        final Set<Integer> dont_stop = new HashSet<>();
        final Set<Integer> dont_pass = new HashSet<>();
        final IProcessesV1 ps_v1 = channel.getRemoteService(IProcessesV1.class);
        if (ps_v1 != null) {
            // Get processes service capabilities
            new LaunchStep() {
                @Override
                void start() throws Exception {
                    ps_v1.getCapabilities(null, new IProcessesV1.DoneGetCapabilities() {
                        @Override
                        public void doneGetCapabilities(IToken token, Exception error, Map<String, Object> properties) {
                            can_terminate_attached = properties != null &&
                                    properties.get("CanTerminateAttached") instanceof Boolean &&
                                    ((Boolean) properties.get("CanTerminateAttached")).booleanValue();
                            done();
                        }
                    });
                }
            };
        }
        // Start the process
        if (!isBlank(debugAdapterConfiguration.getRemoteFile())) {
            new LaunchStep() {
                @Override
                void start() throws Exception {
                    if (env != null) {
                        for (Map.Entry<String, String> e : env.entrySet()) {
                            String key = e.getKey();
                            String val = e.getValue();
                            if (val == null) {
                                process_env.remove(key);
                            } else {
                                process_env.put(key, val);
                            }
                        }
                    }
                    String file = debugAdapterConfiguration.getRemoteFile();
                    if (file == null || file.length() == 0) {
                        channel.terminate(new Exception("Program file does not exist"));
                        return;
                    }
                    IProcesses.DoneStart done = new IProcesses.DoneStart() {
                        public void doneStart(IToken token, final Exception error, IProcesses.ProcessContext process) {
                            if (error != null) {
                                Protocol.sync(new Runnable() {
                                    public void run() {
                                        channel.terminate(error);
                                    }
                                });
                            } else {
                                TCFLaunch.this.process = process;
                                ps.addListener(prs_listener);
                                onAttach(process);
                                done();
                            }
                        }
                    };
                    // clientOut ("Starting: " + file);
                    ProgressEndEventArguments pEndArgs = new ProgressEndEventArguments();
                    pEndArgs.setMessage("Download Completed, Starting: " + file);
                    pEndArgs.setProgressId("1");
                    client.progressEnd(pEndArgs);
                    String cmd = args;
                    String[] args_arr = toArgsArray(file, cmd);
                    if (ps_v1 != null) {
                        Map<String, Object> params = new HashMap<String, Object>();
                        // Current default is debug mode.
                        // if (mode.equals(ILaunchManager.DEBUG_MODE)) {
                        params.put(IProcessesV1.START_ATTACH, true);
                        params.put(IProcessesV1.START_ATTACH_CHILDREN, attach_children);
                        params.put(IProcessesV1.START_STOP_AT_ENTRY, stop_at_entry);
                        params.put(IProcessesV1.START_STOP_AT_MAIN, stop_at_main);
                        if (dont_stop.size() > 0) params.put(IProcessesV1.START_SIG_DONT_STOP, dont_stop);
                        if (dont_pass.size() > 0) params.put(IProcessesV1.START_SIG_DONT_PASS, dont_pass);
                        // }
                        if (use_terminal) params.put(IProcessesV1.START_USE_TERMINAL, true);
                        ps_v1.start(dir, file, args_arr, process_env, params, done);
                    } else {
                        // if psv1 not present.
                        // boolean attach = mode.equals(ILaunchManager.DEBUG_MODE);
                        // Making default attach as false.
                        boolean attach = false;
                        ps.start(dir, file, args_arr, process_env, attach, done);
                    }
                }
            };


            //TODO: For debug mode only, need to revisit this for Run mode.
            // If debug mode.
            new LaunchStep() {
                @Override
                void start() {
                    ps.getSignalList(process.getID(), new IProcesses.DoneGetSignalList() {
                        public void doneGetSignalList(IToken token, Exception error, Collection<Map<String, Object>> list) {
                            if (error != null && attached_processes.get(process.getID()) != null) {
                                LOGGER.info("Can't get process signal list" + error.toString());
                            }
                            process_signals = list;
                            done();
                        }
                    });
                }
            };

            // Set process signal masks
            if (ps_v1 == null && (dont_stop.size() > 0 || dont_pass.size() > 0)) {
                new LaunchStep() {
                    @Override
                    void start() {
                        final HashSet<IToken> cmds = new HashSet<IToken>();
                        final IProcesses.DoneCommand done_set_mask = new IProcesses.DoneCommand() {
                            public void doneCommand(IToken token, Exception error) {
                                cmds.remove(token);
                                if (error != null && attached_processes.size() > 0) channel.terminate(error);
                                else if (cmds.size() == 0) done();
                            }
                        };
                        cmds.add(ps.setSignalMask(process.getID(), dont_stop, dont_pass, done_set_mask));
                        final IRunControl rc = channel.getRemoteService(IRunControl.class);
                        if (rc != null) {
                            final IRunControl.DoneGetChildren done_get_children = new IRunControl.DoneGetChildren() {
                                public void doneGetChildren(IToken token, Exception error, String[] context_ids) {
                                    if (context_ids != null) {
                                        for (String id : context_ids) {
                                            cmds.add(ps.setSignalMask(id, dont_stop, dont_pass, done_set_mask));
                                            cmds.add(rc.getChildren(id, this));
                                        }
                                    }
                                    cmds.remove(token);
                                    if (error != null && attached_processes.size() > 0) channel.terminate(error);
                                    else if (cmds.size() == 0) done();
                                }
                            };
                            cmds.add(rc.getChildren(process.getID(), done_get_children));
                        }
                    }
                };
            }
        }

    }
    public void onAttach(IProcesses.ProcessContext ctx) {
        if (context_filter != null) context_filter.add(ctx.getID());
        attached_processes.put(ctx.getID(), ctx);
        readProcessStreams(ctx);
    }

    public void onDetach(String prs_id) {
        disconnectUnusedStreams();
    }


    private final IProcesses.ProcessesListener prs_listener = new IProcesses.ProcessesListener() {

        public void exited(String process_id, int exit_code) {
            if (process != null && process_id.equals(process.getID())) {
                process_exit_code = exit_code;
                process_exited = true;
            }
            attached_processes.remove(process_id);
        }
    };
    private String[] toArgsArray(String file, String cmd) {
        // Create arguments list from a command line.
        int i = 0;
        int l = cmd.length();
        List<String> arr = new ArrayList<String>();
        arr.add(file);
        for (;;) {
            while (i < l && cmd.charAt(i) == ' ') i++;
            if (i >= l) break;
            String s = null;
            if (cmd.charAt(i) == '"') {
                i++;
                StringBuffer bf = new StringBuffer();
                while (i < l) {
                    char ch = cmd.charAt(i++);
                    if (ch == '"') break;
                    if (ch == '\\' && i < l) ch = cmd.charAt(i++);
                    bf.append(ch);
                }
                s = bf.toString();
            }
            else {
                int i0 = i;
                while (i < l && cmd.charAt(i) != ' ') i++;
                s = cmd.substring(i0, i);
            }
            arr.add(s);
        }
        return arr.toArray(new String[arr.size()]);
    }
    private void mkDirOnRemoteTarget(String remote_dir, final Runnable done){
        final IFileSystem fs = channel.getRemoteService(IFileSystem.class);
        if (fs == null) {
            channel.terminate(new Exception(
                    "Cannot create Directory: target does not provide File System service"));
            return;
        }

        fs.mkdir(remote_dir,null,new DoneMkDir() {

            @Override
            public void doneMkDir(IToken token, FileSystemException error) {
                if (error != null) {
                    TCFLaunch.this.error = new Exception("Cannot create directory", error);
                    clientErr("Cannot create directory " + remote_dir );
                    clientErr(error.toString());
                }
                done.run();
            }
        });
    }
    private void copyFileToRemoteTarget(String local_file, String remote_file, Boolean showEndProgress, final Runnable done) {
        if (local_file == null) {
            channel.terminate(new Exception("Program does not exist"));
            return;
        }
        final IFileSystem fs = channel.getRemoteService(IFileSystem.class);
        if (fs == null) {
            channel.terminate(new Exception(
                    "Cannot download program file: target does not provide File System service"));
            return;
        }
        try {
            final File local_fd = new File(local_file);
            final InputStream inp = new FileInputStream(local_fd);
            final String task_name = "\nDownloading: " + local_fd.getName();
            final String task_name_progress_label = "Downloading: " + local_fd.getName();

            int flags = TCF_O_WRITE | TCF_O_CREAT | TCF_O_TRUNC;
            // Progress monitoring code from eclipse ..
            // if (launch_monitor != null) launch_monitor.subTask(task_name);
            // clientOut(task_name + "\n");
            ProgressStartEventArguments pStartArgs = new ProgressStartEventArguments();
            pStartArgs.setTitle(task_name_progress_label);
            pStartArgs.setCancellable(false);
            pStartArgs.setProgressId("1");
            client.progressStart(pStartArgs);
            fs.open(remote_file, flags, null, new DoneOpen() {

                IFileHandle handle;
                long offset = 0;
                final Set<IToken> cmds = new HashSet<IToken>();
                final byte[] buf = new byte[0x1000];

                public void doneOpen(IToken token, FileSystemException error, IFileHandle handle) {
                    this.handle = handle;
                    if (error != null) {
                        TCFLaunch.this.error = new Exception("Cannot download program file", error);
                        clientErr("Cannot download program file");
                        // fireChanged(); used to notify the launch listeners in eclipse.
                        done.run();
                    }
                    else {
                        write_next();
                    }
                }

                private void write_next() {
                    try {
                        while (cmds.size() < 8) {
                            int rd = inp.read(buf);
                            if (rd < 0) {
                                close();
                                break;
                            }
                            final long kb_done = (offset + rd) / 1024;
                            cmds.add(fs.write(handle, offset, buf, 0, rd, new DoneWrite() {

                                public void doneWrite(IToken token, FileSystemException error) {
                                    cmds.remove(token);
                                    //clientOut(task_name + ", " + kb_done + " KB done\n");
                                    ProgressUpdateEventArguments pUpdateArgs = new ProgressUpdateEventArguments();
                                    pUpdateArgs.setMessage(kb_done + " KB done");
                                    pUpdateArgs.setProgressId("1");
                                    client.progressUpdate(pUpdateArgs);

                                    if (error != null) channel.terminate(error);
                                    else write_next();
                                }
                            }));
                            offset += rd;
                        }
                    }
                    catch (Throwable x) {
                        channel.terminate(x);
                    }
                }

                private void close() {
                    if (cmds.size() > 0) return;
                    try {
                        inp.close();
                        fs.close(handle, new DoneClose() {

                            public void doneClose(IToken token, FileSystemException error) {
                                if (error != null) channel.terminate(error);
                                else done.run();
                            }
                        });
                        if(showEndProgress) {
                            ProgressEndEventArguments pEndArgs = new ProgressEndEventArguments();
                            pEndArgs.setMessage("Download Completed : " + local_file);
                            pEndArgs.setProgressId("1");
                            client.progressEnd(pEndArgs);
                        }
                    }
                    catch (Throwable x) {
                        channel.terminate(x);
                    }
                }
            });
        }
        catch (Throwable x) {
            channel.terminate(x);
        }
    }

    static class FileInfo {
        String name;
        String fullname;
        boolean isDir;
        FileInfo[] children;
        Throwable children_error;
        int index;
        boolean children_pending;
        FileInfo parent;
    }
    public void copyMatchingFilesFromTarget(Map<String, String> filePatterns){
        if(filePatterns.isEmpty()) return;
        String workingDirectory = debugAdapterConfiguration.getWorkingDirectory();
        if(isBlank(workingDirectory)){
            workingDirectory="/mnt";
        }
        final IFileSystem fs = channel.getRemoteService(IFileSystem.class);
        if (fs == null) {
            channel.terminate(new Exception(
                    "Cannot Copy the file: target does not provide File System service"));
            return;
        }
        String finalWorkingDirectory = workingDirectory;
        final List<FileInfo> fileInfos = new ArrayList<FileInfo>();
        String finalWorkingDirectory1 = workingDirectory;
        fs.opendir(workingDirectory, new IFileSystem.DoneOpen() {

            @Override
            public void doneOpen(IToken token, FileSystemException error, IFileHandle handle) {
                if (error != null) {
                    channel.terminate(new Exception(
                            "Cannot read directory: " + finalWorkingDirectory));
                    return;
                }
                fs.readdir(handle, new IFileSystem.DoneReadDir() {
                    public void doneReadDir(IToken token, FileSystemException error, DirEntry[] entries, boolean eof) {
                        if (entries != null) {
                            for (DirEntry entry : entries) {
                                FileInfo info = new FileInfo();
                                info.parent = new FileInfo();
                                info.parent.name = finalWorkingDirectory1;
                                info.parent.fullname = finalWorkingDirectory1;
                                info.name = entry.filename;
                                info.fullname = entry.longname != null ? entry.longname : (Paths.get(finalWorkingDirectory1,info.name).toString());
                                info.isDir = entry.attrs != null ? entry.attrs.isDirectory() : false;
//                                if (!fDirectoriesOnly || info.isDir) {
//                                    fileInfos.add(info);
//                                }
                                fileInfos.add(info);
                            }
                        }
                        if (error != null || eof) {
                            fs.close(handle, new IFileSystem.DoneClose() {
                                public void doneClose(IToken token, FileSystemException error) {
                                    // ignore error
                                }
                            });
                            int size = fileInfos.size();
                            if (size == 0 && error != null) {
                                doneLoadChildren(error, null,finalWorkingDirectory1,filePatterns);
                            } else {
                                doneLoadChildren(null, fileInfos,finalWorkingDirectory1,filePatterns);
                            }
                        } else {
                            fs.readdir(handle, this);
                        }
                    }
                });
            }
        });
    }

    private void doneLoadChildren(FileSystemException object, List<FileInfo> fileInfoList, String workingDirectory, Map<String, String> filePatterns) {
        for(FileInfo fileInfo : fileInfoList ){
            if(!fileInfo.isDir) {
                Optional<String> first = filePatterns.keySet().stream().filter(pattern -> fileInfo.name.contains(pattern)).findFirst();
                if(first.isPresent()){
                    String destinationDirectory = filePatterns.get(first.get());
                    Path localFilePath = Paths.get(destinationDirectory,fileInfo.name);
                    copyFileFromRemoteTarget(fileInfo.fullname, localFilePath.toString());
                }
            }
        }
    }

    public  void copyFileFromRemoteTarget(String remote_file,String local_file) {
        if (remote_file == null) {
            channel.terminate(new Exception("File does not exist"));
            return;
        }
        final IFileSystem fs = channel.getRemoteService(IFileSystem.class);
        if (fs == null) {
            channel.terminate(new Exception(
                    "Cannot Copy the file: target does not provide File System service"));
            return;
        }
        try {
            if(! Paths.get(local_file).getParent().toFile().exists() ){
                Paths.get(local_file).getParent().toFile().mkdirs();
            }
            final File local_fd = new File(local_file);
            final OutputStream outp = new FileOutputStream(local_fd);
            final String task_name = "\nDownloading: " + local_fd.getName();
            final String task_name_progress_label = "Copying: " + local_fd.getName();

            int flags = TCF_O_READ;
            ProgressStartEventArguments pStartArgs = new ProgressStartEventArguments();
            pStartArgs.setTitle(task_name_progress_label);
            pStartArgs.setCancellable(false);
            pStartArgs.setProgressId("1");
            client.progressStart(pStartArgs);
            fs.open(remote_file, flags, null, new DoneOpen() {

                IFileHandle handle;
                long offset = 0;
                final Set<IToken> cmds = new HashSet<IToken>();
                final byte[] buf = new byte[0x1000];

                public void doneOpen(IToken token, FileSystemException error, IFileHandle handle) {
                    this.handle = handle;
                    if (error != null) {
                        TCFLaunch.this.error = new Exception("Cannot Read  file", error);
                        clientErr("Cannot read file :" + remote_file);
                        // fireChanged(); used to notify the launch listeners in eclipse.
                    }
                    else {
                        read_next();
                    }
                }

                private void read_next() {
                    try {
                        while (cmds.size() < 1) {
                            final long kb_done = (offset + 0x1000) / 1024;
                            cmds.add(fs.read(handle, offset,  0x1000,  new DoneRead() {
                                @Override
                                public void doneRead(IToken token, FileSystemException error, byte[] data, boolean eof) {
                                    cmds.remove(token);
                                    if (error != null) channel.terminate(error);
                                    else {
                                        if(data.length != 0 ) {
                                            try {
                                                outp.write(data);
                                            } catch (IOException e) {
                                                e.printStackTrace();
                                            }
                                        }
                                        ProgressUpdateEventArguments pUpdateArgs = new ProgressUpdateEventArguments();
                                        pUpdateArgs.setMessage(kb_done + " KB done");
                                        pUpdateArgs.setProgressId("1");
                                        client.progressUpdate(pUpdateArgs);
                                        if(eof){
                                            try {
                                                outp.close();
                                                close();
                                            } catch (IOException e) {
                                                e.printStackTrace();
                                            }
                                        }else {
                                            offset += data.length;
                                            read_next();
                                        }
                                    }
                                }
                            }));
                        }
                    }
                    catch (Throwable x) {
                        channel.terminate(x);
                    }
                }

                private void close() {
                    if (cmds.size() > 0) return;
                    try {
                        outp.close();
                        fs.close(handle, new DoneClose() {

                            public void doneClose(IToken token, FileSystemException error) {
                                if (error != null) channel.terminate(error);
                            }
                        });
                        ProgressEndEventArguments pEndArgs = new ProgressEndEventArguments();
                        pEndArgs.setMessage("Copy Completed : " + remote_file);
                        pEndArgs.setProgressId("1");
                        client.progressEnd(pEndArgs);
                    }
                    catch (Throwable x) {
                        channel.terminate(x);
                    }
                }
            });
        }
        catch (Throwable x) {
            channel.terminate(x);
        }
    }

    private void getMemMaps(Map<String,ArrayList<IMemoryMap.MemoryRegion>> maps) throws Exception {
        ArrayList<IMemoryMap.MemoryRegion> l = new ArrayList<>();
        //TODO: Need to see, how to get this data.
        String id = "P562";
        maps.put(id, l);
        Map<String,Object> map = new HashMap<String,Object>();
        map.put("FileName",debugAdapterConfiguration.getApplicationFile());
        map.put("Flags", 0);
        map.put("Id","P563");
        map.put("UUID","6c9b3f51-bdfd-415f-bb0b-8543a045705f");
        l.add(new TCFMemoryRegion(map));

    }

    private void downloadMemoryMaps(final Runnable done) throws Exception {
        final IMemoryMap mmap = channel.getRemoteService(IMemoryMap.class);
        if (mmap == null) {
            done.run();
            return;
        }
        final HashMap<String,ArrayList<IMemoryMap.MemoryRegion>> maps = new HashMap<String,ArrayList<IMemoryMap.MemoryRegion>>();
        getMemMaps(maps);
        final HashSet<IToken> cmds = new HashSet<IToken>(); // Pending commands
        final Runnable done_all = new Runnable() {
            boolean launch_done;
            public void run() {
                if (launch_done) return;
                done.run();
                launch_done = true;
            }
        };
        final IMemoryMap.DoneSet done_set_mmap = new IMemoryMap.DoneSet() {
            public void doneSet(IToken token, Exception error) {
                assert cmds.contains(token);
                cmds.remove(token);
                if (error != null) {
                    LOGGER.info("Cannot update context memory map" +  error);
                    // Activator.log("Cannot update context memory map", error);
                }
                if (cmds.isEmpty()) done_all.run();
            }
        };
        for (String id : maps.keySet()) {
            ArrayList<IMemoryMap.MemoryRegion> map = maps.get(id);
            TCFMemoryRegion[] arr = map.toArray(new TCFMemoryRegion[map.size()]);
            cmds.add(mmap.set(id, arr, done_set_mmap));
        }
        update_memory_maps = new Runnable() {
            public void run() {
                try {
                    Set<String> set = new HashSet<String>(maps.keySet());
                    maps.clear();
                    getMemMaps(maps);
                    for (String id : maps.keySet()) {
                        ArrayList<IMemoryMap.MemoryRegion> map = maps.get(id);
                        TCFMemoryRegion[] arr = map.toArray(new TCFMemoryRegion[map.size()]);
                        cmds.add(mmap.set(id, arr, done_set_mmap));
                    }
                    for (String id : set) {
                        if (maps.get(id) != null) continue;
                        cmds.add(mmap.set(id, null, done_set_mmap));
                    }
                }
                catch (Throwable x) {
                    channel.terminate(x);
                }
            }
        };
        if (cmds.isEmpty()) done_all.run();

    }
    private void readPathMapConfiguration(TCFDebugAdapterConfiguration debugAdapterConfiguration) {
        host_path_map = new ArrayList<IPathMap.PathMapRule>();
        host_path_map.addAll(extractPathMapAttribute(debugAdapterConfiguration));


    }
    public void onLastContextRemoved() {
        boolean disconnectOnExit = true;
        // boolean disconnectOnExit =  debugAdapterConfiguration.getDisconnectOnExit();
        try {
            if (process != null && disconnectOnExit ) {
                last_context_exited = true;
                closeChannel();
            }
        }
        catch (Throwable e) {
            clientErr("Cannot access launch configuration" +  e.toString());
        }
    }
    public boolean isConnecting() {
        return connecting;
    }

    public boolean isConnected() {
        return channel != null && !connecting && !disconnected;
    }

    public void closeChannel() {
        assert Protocol.isDispatchThread();
        if (channel == null) return;
        if (channel.getState() == IChannel.STATE_CLOSED) return;
        if (disconnecting) return;
        disconnecting = true;
        final Set<IToken> cmds = new HashSet<IToken>();
        if (channel.getState() == IChannel.STATE_OPEN) {
            for (final IProcesses.ProcessContext ctx : attached_processes.values()) {
                IProcesses.DoneCommand prs_done_cmd = new IProcesses.DoneCommand() {
                    public void doneCommand(IToken token, Exception error) {
                        cmds.remove(token);
                        if (!attached_processes.containsKey(ctx.getID())) error = null;
                        if (error != null) channel.terminate(error);
                        else if (cmds.isEmpty()) channel.close();
                    }
                };
                if (!can_terminate_attached) cmds.add(ctx.detach(prs_done_cmd));
                cmds.add(ctx.terminate(prs_done_cmd));
            }
            IStreams streams = getService(IStreams.class);
            IStreams.DoneDisconnect done_disconnect = new IStreams.DoneDisconnect() {
                public void doneDisconnect(IToken token, Exception error) {
                    cmds.remove(token);
                    if (error != null) channel.terminate(error);
                    else if (cmds.isEmpty()) channel.close();
                }
            };
            for (String id : process_stream_ids.keySet()) {
                cmds.add(streams.disconnect(id, done_disconnect));
            }
            for (String id : uart_rx_stream_ids.keySet()) {
                cmds.add(streams.disconnect(id, done_disconnect));
            }
            for (String id : uart_tx_stream_ids.keySet()) {
                cmds.add(streams.disconnect(id, done_disconnect));
            }
            process_stream_ids.clear();
            uart_rx_stream_ids.clear();
            uart_tx_stream_ids.clear();
            if (dprintf_stream_id != null) {
                disconnected_stream_ids.add(dprintf_stream_id);
                cmds.add(streams.disconnect(dprintf_stream_id, done_disconnect));
                dprintf_stream_id = null;
            }
        }
        if (cmds.isEmpty()) channel.close();
    }


    public boolean isExited() {
        return last_context_exited;
    }

    public boolean isProcessExited() {
        return process_exited;
    }

    public int getExitCode() {
        return process_exit_code;
    }

    private Collection<? extends IPathMap.PathMapRule> extractPathMapAttribute(TCFDebugAdapterConfiguration debugAdapterConfiguration) {
        ArrayList<IPathMap.PathMapRule> map = new ArrayList<IPathMap.PathMapRule>();
        IPathMap.PathMapRule e = new PathMapRule(new LinkedHashMap<String,Object>());
        if(!isBlank(debugAdapterConfiguration.getRemoteFile())) {
            e.getProperties().put("Source", debugAdapterConfiguration.getRemoteFile());
            e.getProperties().put("Destination", debugAdapterConfiguration.getApplicationFile());
            map.add(e);
        }
        List<ImmutablePair<String, String>> pathMapEntriesList = debugAdapterConfiguration.getPathMapEntriesList();
        for(ImmutablePair<String,String> pathMapEntry : pathMapEntriesList){
            IPathMap.PathMapRule entry = new PathMapRule(new LinkedHashMap<String,Object>());
            entry.getProperties().put("Source", pathMapEntry.getLeft());
            entry.getProperties().put("Destination", pathMapEntry.getRight());
            map.add(entry);
        }
        return map;
    }

    private Map<String, String> createAttributes() {
        Map<String, String> attrs = new HashMap<String, String>();
        attrs.put(IPeer.ATTR_ID, "RigelPeer");
        attrs.put(IPeer.ATTR_SERVICE_MANGER_ID, ServiceManager.getID());
        attrs.put(IPeer.ATTR_AGENT_ID, Protocol.getAgentID());
        attrs.put(IPeer.ATTR_NAME, "Rigel Peer");
        attrs.put(IPeer.ATTR_OS_NAME, System.getProperty("os.name"));
        attrs.put(IPeer.ATTR_TRANSPORT_NAME, "TCP");
        attrs.put(IPeer.ATTR_IP_HOST, debugAdapterConfiguration.getHostName());
        attrs.put(IPeer.ATTR_IP_PORT, debugAdapterConfiguration.getPort());
        return attrs;
    }
    public void openUartStreams(final String ctx_id, Map<String,Object> uart_props) {
        assert Protocol.isDispatchThread();
        if (uart_props == null) return;
        IStreams streams = getService(IStreams.class);
        if (streams == null) return;
        final String rx_id = (String)uart_props.get("RXStreamID");
        if (rx_id != null && uart_rx_stream_ids.get(rx_id) == null) {
            streams.connect(rx_id, new IStreams.DoneConnect() {
                @Override
                public void doneConnect(IToken token, Exception error) {
                    if (uart_rx_stream_ids.get(rx_id) != null) return;
                    uart_rx_stream_ids.put(rx_id, ctx_id);
                    if (error == null) return;
                    for (LaunchListener l : getListeners()) l.onProcessStreamError(TCFLaunch.this, ctx_id, 0, error, 0);
                }
            });
        }
        final String tx_id = (String)uart_props.get("TXStreamID");
        if (tx_id != null && uart_tx_stream_ids.get(tx_id) == null) {
            streams.connect(tx_id, new IStreams.DoneConnect() {
                @Override
                public void doneConnect(IToken token, Exception error) {
                    if (uart_tx_stream_ids.get(tx_id) != null) return;
                    uart_tx_stream_ids.put(tx_id, ctx_id);
                    if (error == null) {
                        readStream(ctx_id, tx_id, 0);
                        return;
                    }
                    for (LaunchListener l : getListeners()) l.onProcessStreamError(TCFLaunch.this, ctx_id, 0, error, 0);
                }
            });
        }
    }

    protected void clientOutput(String output, String category) {
        if (this.client != null) {
            OutputEventArguments arguments = new OutputEventArguments();
            arguments.setCategory(category);
            arguments.setOutput(output);
            this.client.output(arguments);
        }
    }

    protected void clientErr(String output) {
        clientOutput(output, OutputEventArgumentsCategory.STDERR);
    }

    protected void clientOut(String output) {
        clientOutput(output, "stdout");
    }

    protected void clientStop(){
        StoppedEventArguments args = new StoppedEventArguments();
        // args.setDescription("Stopped at main");
        // args.setReason("Breakpoint hit at main");
        args.setAllThreadsStopped(true);
        client.stopped(args);
    }

    /**
     * Get current target breakpoints status information.
     * @return status information object
     */
    public TCFBreakpointsStatus getBreakpointsStatus() {
        return breakpoints_status;
    }

    public TCFBreakpointHandler getBreakpointHandler() {
        return breakpointHandler;
    }
    /**
     * Get context filer of the launch.
     * By default, TCF debugger shows all remote debug contexts.
     * Context filter is used to hide unwanted contexts to reduce UI clutter.
     * @return context filter.
     */
    public Set<String> getContextFilter() {
        return context_filter;
    }

    public void dispose() {
        bp_status_listener.dispose();
    }
}
