/*******************************************************************************
 * Copyright (c) 2007-2018 Wind River Systems, Inc. and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License 2.0
 * which accompanies this distribution, and is available at
 * https://www.eclipse.org/legal/epl-2.0/
 *
 * Contributors:
 *     Wind River Systems - initial API and implementation
 *******************************************************************************/

/*******************************************************************************
 * Modified by Xilinx,Inc
 * Copyright 2022 Xilinx, Inc. All Rights Reserved. Modifications
 * are made available under the terms of the Eclipse Public License 2.0
 * which accompanies this distribution, and is available at
 * https://www.eclipse.org/legal/epl-2.0/
 *
 *******************************************************************************/

package org.eclipse.tcf.internal.dap;

import org.apache.commons.lang3.tuple.ImmutablePair;
import org.apache.commons.lang3.tuple.ImmutableTriple;
import org.eclipse.lsp4j.debug.*;
import org.eclipse.tcf.commands.*;
import org.eclipse.tcf.core.Command;
import org.eclipse.tcf.core.ILaunch;
import org.eclipse.tcf.dap.ITCFModel;
import org.eclipse.tcf.dap.ITCFObject;
import org.eclipse.tcf.dap.ITCFPresentationProvider;
import org.eclipse.tcf.internal.dap.actions.TCFAction;
import org.eclipse.tcf.internal.dap.command.*;
import org.eclipse.tcf.protocol.*;
import org.eclipse.tcf.services.*;
import org.eclipse.tcf.util.TCFDataCache;

import java.math.BigInteger;
import java.net.UnknownHostException;
import java.util.*;
import java.util.concurrent.CompletableFuture;
import java.util.logging.Logger;

/**
 * TCFModel represents remote target state as it is known to host.
 * The main job of the model is caching remote data,
 * keeping the cache in a coherent state,
 * and feeding UI with up-to-date data.
 */
public class TCFModel implements ITCFModel {
    protected static final Logger LOGGER = Logger.getLogger(TCFModel.class.getSimpleName());
    public static final String ID_EXPRESSION_HOVER = "hover";
    public static final String ID_VARIABLES = "variables";
    public static final String ID_EXPRESSIONS = "expressions";
    private IChannel channel;
    private TCFNodeLaunch launch_node;
    /** Selection reason in the Debug view - context added */
    public static final String SELECT_ADDED = "Added";
    int sourceRequestNumber = 100;
    Map<Integer, IDisassembly.IDisassemblyLine[]> disassembleMap = new HashMap<>();

    private final ExpressionManager expr_manager = new ExpressionManager();
    private boolean disposed;
    private final TCFLaunch launch;
    private final Map<String,String> cast_to_type_map = new HashMap<String,String>();
    private final Map<Class<?>,Object> adapters = new HashMap<Class<?>,Object>();
    private final Map<String,String> action_results = new HashMap<String,String>();
    private final HashMap<String, TCFAction> active_actions = new HashMap<String,TCFAction>();
    private final Map<String, TCFNode> id2node = new HashMap<String,TCFNode>();
    private final Map<String,Object> context_map = new HashMap<String,Object>();
    private InitialSuspendTrigger initial_suspend_trigger;
    // Removing final, making a bit change.
    List<ITCFPresentationProvider> view_request_listeners = new ArrayList<>();
    private boolean stack_frames_limit_enabled;
    private boolean filter_variants_by_discriminant;
    private int stack_frames_limit_value;
    private boolean show_function_arg_names;
    private boolean show_function_arg_values;
    private long min_view_updates_interval;
    private boolean view_updates_throttle_enabled;
    private boolean channel_throttle_enabled;


    private int auto_disconnect_generation;
    private boolean qualified_type_names_enabled;


    TCFModel(final TCFLaunch launch) {
        this.launch = launch;
        adapters.put(ILaunch.class, launch);
        adapters.put(IStepOverHandler.class, new StepOverCommand(this));
        adapters.put(IStepIntoHandler.class, new StepIntoCommand(this));
        adapters.put(IStepReturnHandler.class, new StepReturnCommand(this));
        adapters.put(ITerminateHandler.class, new TerminateCommand(this));
        adapters.put(IResumeHandler.class, new ResumeCommand(this));
        adapters.put(ISuspendHandler.class, new SuspendCommand(this));
    }
    public ExpressionManager getExpressionManager(){
        return expr_manager;
    }
    /**
     * Add an adapter for given type.
     *
     * @param adapterType  the type the adapter implements
     * @param adapter  the adapter implementing <code>adapterType</code>
     */
    public void setAdapter(Class<?> adapterType, Object adapter) {
        synchronized (adapters) {
            adapters.put(adapterType, adapter);
        }
    }

    @SuppressWarnings("rawtypes")
    public Object getAdapter(final Class adapter, final TCFNode node) {
        synchronized (adapters) {
            Object o = adapters.get(adapter);
            if (o != null) return o;
        }
        // TODO:
//        if (adapter == IMemoryBlockRetrieval.class || adapter == IMemoryBlockRetrievalExtension.class) {
//            return new TCFTask<Object>() {
//                public void run() {
//                    Object o = null;
//                    TCFDataCache<TCFNodeExecContext> cache = searchMemoryContext(node);
//                    if (cache != null) {
//                        if (!cache.validate(this)) return;
//                        if (cache.getData() != null) {
//                            TCFNodeExecContext ctx = cache.getData();
//                            if (!ctx.getMemoryContext().validate(this)) return;
//                            if (ctx.getMemoryContext().getError() == null) {
//                                o = getMemoryBlockRetrieval(ctx.id);
//                            }
//                        }
//                    }
//                    if (o == null) o = mem_not_supported;
//                    assert adapter.isInstance(o);
//                    done(o);
//                }
//            }.getE();
//        }
        return null;
    }
    public long getMinViewUpdatesInterval() {
        return min_view_updates_interval;
    }

    public boolean getViewUpdatesThrottleEnabled() {
        return view_updates_throttle_enabled;
    }

    public boolean getChannelThrottleEnabled() {
        return channel_throttle_enabled;
    }


    /**
     * Get a data cache that contains properties of a symbol.
     * New cache object is created if it does not exist yet.
     * @param sym_id - the symbol ID.
     * @return data cache object.
     */
    public TCFDataCache<ISymbols.Symbol> getSymbolInfoCache(final String sym_id) {
        if (sym_id == null) return null;
        TCFNodeSymbol n = (TCFNodeSymbol)getNode(sym_id);
        if (n == null) n = new TCFNodeSymbol(launch_node, sym_id);
        return n.getContext();
    }

    /**
     * @return debug model launch object.
     */
    public TCFLaunch getLaunch() {
        return launch;
    }
    void addNode(String id, TCFNode node) {
        assert id != null;
        assert Protocol.isDispatchThread();
        assert id2node.get(id) == null;
        assert launch_node != null;
        assert !node.isDisposed();
        id2node.put(id, node);
    }
    public boolean getShowFunctionArgNames() {
        return show_function_arg_names;
    }

    public boolean getShowFunctionArgValues() {
        return show_function_arg_values;
    }



    Map<String,Object> getContextMap() {
        return context_map;
    }

    public void getValidRunContext(TCFNodeExecContext threadNodeExecContext, CompletableFuture<IRunControl.RunControlContext> runContextFuture) {
        Protocol.invokeLater(25, new Runnable() {
            public void run() {
                TCFDataCache<IRunControl.RunControlContext> runContext = threadNodeExecContext.getRunContext();
                if (!runContext.validate(this)) return;
                if (!runContext.isDisposed()) {
                    runContextFuture.complete(runContext.getData());
                } else {
                    runContextFuture.complete(null);
                }
            }
        });

    }
    public void getValidNodeExpression(CompletableFuture<TCFChildrenLocalVariables> childrenFuture) {
        Protocol.invokeLater(25, new Runnable() {
                    public void run() {

                    }
                }
        );
    }

    public void getValidChildren(TCFNodeExecContext threadNodeExecContext, CompletableFuture<TCFChildren> childrenFuture) {
        Protocol.invokeLater(25, new Runnable() {
            public void run() {
                TCFChildren children = threadNodeExecContext.getChildren();
                if (!children.validate(this)) return;
                if (!children.isDisposed()) {
                    childrenFuture.complete(children);
                } else {
                    childrenFuture.complete(null);
                }
            }
        });

    }

    public ImmutablePair<Integer,Integer> getVariablesCount(TCFNodeStackFrame nodeStackFrame, Runnable done) {
        assert Protocol.isDispatchThread();
        int indexedVarCount = 0;
        int namedVarCount = 0;
        if(nodeStackFrame == null || nodeStackFrame.isDisposed() ){
            return ImmutablePair.of(0,0);
        }
        TCFChildren var_children = nodeStackFrame.getChildren(ID_VARIABLES);
        if (!var_children.validate(done)) return null;

        Map<String, ITCFObject> children = var_children.getChildren();
        for (Map.Entry<String, ITCFObject> e : children.entrySet()) {
            ITCFObject value = e.getValue();
            TCFNodeExpression nodeExpression = (TCFNodeExpression) value;
            TCFDataCache<ISymbols.Symbol> type = nodeExpression.getType();
            if(!type.validate(done)) return null;
            ISymbols.Symbol typeData = type.getData();
            if(typeData != null ) {
                if (typeData.getTypeClass() == ISymbols.TypeClass.array) {
                    indexedVarCount++;
                } else
                    namedVarCount++;
            }
        }
        return ImmutablePair.of(namedVarCount,indexedVarCount);
    }

    public ImmutablePair<String, EvaluateResponse> getHoverExprEval(Object object, String expr,
                                                                    Runnable done) {

        if (object instanceof TCFNodeStackFrame) {
            TCFNodeStackFrame node =  (TCFNodeStackFrame) object;
            TCFChildren cache = node.getModel().getHoverExpressionCache(node, expr);
            if (!cache.validate(done)) return null;
            Map<String, TCFNode> nodes = cache.getData();
            if (nodes != null) {
                boolean ok = false;
                for (TCFNode n : nodes.values()) {
                    TCFNodeExpression node_expr = (TCFNodeExpression) n;
                    // if (running) expr.update(this);
                    node_expr.update(done);
                    TCFDataCache<IExpressions.Value> value = node_expr.getValue();
                    if (!value.validate(done)) return null;
                    if (value.getData() != null) ok = true;
                }
                if (ok) {
                    TCFChildren children = node.getChildren(ID_EXPRESSION_HOVER);
                    if(!children.validate(done)) return null;
                    if(children instanceof  TCFChildrenHoverExpressions ){
                        TCFChildrenHoverExpressions children_hover_expr = (TCFChildrenHoverExpressions) children;
                        Map<String, TCFNode> data = children_hover_expr.getData();
                        TCFNodeExpression nodeExpression = (TCFNodeExpression)( new ArrayList<>(data.values()).get(0));
                        TCFDataCache<ISymbols.Symbol> type = nodeExpression.getType();
                        if(!type.validate(done)) return null;
                        ISymbols.Symbol typeData = type.getData();
                        EvaluateResponse eval_resp = new EvaluateResponse();

                        TCFDataCache<String> varExprText = nodeExpression.getExpressionText();
                        if (!varExprText.validate(done)) return null;
                        String varExprTextData = varExprText.getData();
                        TCFDataCache<String> typeName = nodeExpression.getTypeName();
                        if (!typeName.validate(done)) return null;
                        String typeNameData = typeName.getData();
                        TCFDataCache<String> baseText = nodeExpression.getBaseText();
                        if(!baseText.validate(done)) return null;
                        String baseTextData = baseText.getData();
                        TCFNode.TCFData<IExpressions.Expression> varExpression = nodeExpression.getVarExpression();
                        if(!varExpression.validate(done)) return null;
                        IExpressions.Expression data_expr = varExpression.getData();
                        TCFDataCache<IExpressions.Value> exprValue = nodeExpression.getValue();
                        if(!exprValue.validate(done)) return null;
                        IExpressions.Value exprValueData = exprValue.getData();

                        TCFDataCache<String> valueString = nodeExpression.getValueString();
                        if (!valueString.validate(done)) return null;
                        String valueStringData = valueString.getData();

                        eval_resp.setType(typeNameData);
                        eval_resp.setResult(valueStringData);
                        String nodeExpressionID="";

                        if( (typeData!=null) &&  ( type.getData().getTypeClass() == ISymbols.TypeClass.pointer ||
                                type.getData().getTypeClass() == ISymbols.TypeClass.array ||
                                type.getData().getTypeClass() == ISymbols.TypeClass.composite) ){


                            TCFChildrenSubExpressions childrenSubExpressions = nodeExpression.getChildrenSubExpressions();
                            if (!childrenSubExpressions.validate(done)) return null;
                            if (childrenSubExpressions.getData().size() > 0) {
                                nodeExpressionID = nodeExpression.getID();
                                eval_resp.setNamedVariables(childrenSubExpressions.getData().size());
                                eval_resp.setVariablesReference( Math.abs(nodeExpressionID.hashCode()));
                            }
                        }
                        return ImmutablePair.of(nodeExpressionID,eval_resp);
                    }
                }else{
                    return (ImmutablePair.of("",new EvaluateResponse()));
                }
            }
        }
        return (ImmutablePair.of("",new EvaluateResponse()));
    }
    public Boolean setExprValue(Object object, String expr, String valueToSet,
                                                                   Runnable done, CompletableFuture<Boolean> setFuture) {

        if (object instanceof TCFNodeStackFrame) {
            TCFNodeStackFrame node =  (TCFNodeStackFrame) object;
            TCFChildren cache = node.getModel().getHoverExpressionCache(node, expr);
            if (!cache.validate(done)) return null;
            Map<String, TCFNode> nodes = cache.getData();
            if (nodes != null) {
                boolean ok = false;
                for (TCFNode n : nodes.values()) {
                    TCFNodeExpression node_expr = (TCFNodeExpression) n;
                    // if (running) expr.update(this);
                    node_expr.update(done);
                    TCFDataCache<IExpressions.Value> value = node_expr.getValue();
                    if (!value.validate(done)) return null;
                    if (value.getData() != null) ok = true;
                }
                if (ok) {
                    TCFChildren children = node.getChildren(ID_EXPRESSION_HOVER);
                    if(!children.validate(done)) return null;
                    if(children instanceof  TCFChildrenHoverExpressions ){
                        TCFChildrenHoverExpressions children_hover_expr = (TCFChildrenHoverExpressions) children;
                        Map<String, TCFNode> data = children_hover_expr.getData();
                        TCFNodeExpression nodeExpression = (TCFNodeExpression)( new ArrayList<>(data.values()).get(0));
                        try {
                            Boolean status = nodeExpression.modify(valueToSet, done, setFuture);
                        } catch (Exception exception) {
                            exception.printStackTrace();
                        }
//                        TCFDataCache<ISymbols.Symbol> type = nodeExpression.getType();
//                        if(!type.validate(done)) return null;
//                        ISymbols.Symbol typeData = type.getData();
//                        EvaluateResponse eval_resp = new EvaluateResponse();
//
//                        TCFDataCache<String> varExprText = nodeExpression.getExpressionText();
//                        if (!varExprText.validate(done)) return null;
//                        String varExprTextData = varExprText.getData();
//                        TCFDataCache<String> typeName = nodeExpression.getTypeName();
//                        if (!typeName.validate(done)) return null;
//                        String typeNameData = typeName.getData();
//                        TCFDataCache<String> baseText = nodeExpression.getBaseText();
//                        if(!baseText.validate(done)) return null;
//                        String baseTextData = baseText.getData();
//                        TCFNode.TCFData<IExpressions.Expression> varExpression = nodeExpression.getVarExpression();
//                        if(!varExpression.validate(done)) return null;
//                        IExpressions.Expression data_expr = varExpression.getData();
//                        TCFDataCache<IExpressions.Value> exprValue = nodeExpression.getValue();
//                        if(!exprValue.validate(done)) return null;
//                        IExpressions.Value exprValueData = exprValue.getData();
//
//                        TCFDataCache<String> valueString = nodeExpression.getValueString();
//                        if (!valueString.validate(done)) return null;
//                        String valueStringData = valueString.getData();
//
//                        eval_resp.setType(typeNameData);
//                        eval_resp.setResult(valueStringData);
//                        String nodeExpressionID="";
//
//                        if( (typeData!=null) &&  ( type.getData().getTypeClass() == ISymbols.TypeClass.pointer ||
//                                type.getData().getTypeClass() == ISymbols.TypeClass.array ||
//                                type.getData().getTypeClass() == ISymbols.TypeClass.composite) ){
//
//
//                            TCFChildrenSubExpressions childrenSubExpressions = nodeExpression.getChildrenSubExpressions();
//                            if (!childrenSubExpressions.validate(done)) return null;
//                            if (childrenSubExpressions.getData().size() > 0) {
//                                nodeExpressionID = nodeExpression.getID();
//                                eval_resp.setNamedVariables(childrenSubExpressions.getData().size());
//                                eval_resp.setVariablesReference( Math.abs(nodeExpressionID.hashCode()));
//                            }
//                        }
//                        return ImmutablePair.of(nodeExpressionID,eval_resp);
                    return true;
                    }
                }else{
                    if(!setFuture.isDone()) setFuture.complete(false);
                    return false;
                }
            }else{
                if(!setFuture.isDone()) setFuture.complete(false);
            }
        }else{
            if(!setFuture.isDone()) setFuture.complete(false);
        }
        return false;
    }
    public List<ImmutableTriple<String, Variable,Boolean>> getVariables(Object object, VariablesArgumentsFilter filter, Runnable done) {

        List<ImmutableTriple<String,Variable,Boolean>> variableList = new LinkedList<>();
        TCFChildren var_children;
        if (object instanceof TCFNodeStackFrame) {
            TCFNodeStackFrame nodeStackFrame = (TCFNodeStackFrame) object;
            var_children = nodeStackFrame.getChildren(ID_VARIABLES);
            if (!var_children.validate(done)) return null;
        }else if( object instanceof TCFNodeExpression ) {
            TCFChildrenSubExpressions childrenSubExpressions = ((TCFNodeExpression) object).getChildrenSubExpressions();
            if(!childrenSubExpressions.validate(done)) return null;
            var_children = childrenSubExpressions;
        }else {
            return (variableList);
        }
            Map<String, ITCFObject> children = var_children.getChildren();
            Map<String, ITCFObject> treeMap = new TreeMap<>(children);
            for (Map.Entry<String, ITCFObject> e : treeMap.entrySet()) {
                ITCFObject value = e.getValue();
                TCFNodeExpression nodeExpression = (TCFNodeExpression) value;
                TCFDataCache<ISymbols.Symbol> type = nodeExpression.getType();
                if (!type.validate(done )) return null;
                ISymbols.Symbol typeData = type.getData();
                if(typeData == null ) continue;
                if((filter == VariablesArgumentsFilter.NAMED) &&
                        (type.getData().getTypeClass() == ISymbols.TypeClass.array)) {
                    continue;
                }
                if((filter == VariablesArgumentsFilter.INDEXED) &&
                        (type.getData().getTypeClass() != ISymbols.TypeClass.array)) {
                    continue;
                }
                    Variable variable = new Variable();
                    TCFDataCache<String> varExprText = nodeExpression.getExpressionText();
                    if (!varExprText.validate(done)) return null;
                    String varExprTextData = varExprText.getData();
                    TCFDataCache<String> typeName = nodeExpression.getTypeName();
                    if (!typeName.validate(done)) return null;
                    String typeNameData = typeName.getData();
                    TCFDataCache<String> baseText = nodeExpression.getBaseText();
                if(!baseText.validate(done)) return null;
                String baseTextData = baseText.getData();

                TCFNode.TCFData<IExpressions.Expression> varExpression = nodeExpression.getVarExpression();
                if(!varExpression.validate(done)) return null;
                IExpressions.Expression data = varExpression.getData();
                TCFDataCache<IExpressions.Value> exprValue = nodeExpression.getValue();
                if(!exprValue.validate(done)) return null;
                IExpressions.Value exprValueData = exprValue.getData();
                TCFDataCache<String> valueString = nodeExpression.getValueString();
                if (!valueString.validate(done)) return null;
                String valueStringData = valueString.getData();
                variable.setName(varExprTextData);
                variable.setType(typeNameData);
                variable.setValue(valueStringData);
                String nodeExpressionID="";
                Boolean isComposite = false;
                nodeExpressionID = nodeExpression.getID();

                if(type.getData().getTypeClass() == ISymbols.TypeClass.pointer ||
                            type.getData().getTypeClass() == ISymbols.TypeClass.array ||
                            type.getData().getTypeClass() == ISymbols.TypeClass.composite){

                        TCFChildrenSubExpressions childrenSubExpressions = nodeExpression.getChildrenSubExpressions();
                        if (!childrenSubExpressions.validate(done)) return null;
                        if (childrenSubExpressions.getData().size() > 0) {
                            variable.setNamedVariables(childrenSubExpressions.getData().size());
                            variable.setVariablesReference( Math.abs(nodeExpressionID.hashCode()));
                            isComposite = true;
                        }
                    }
                    variableList.add(ImmutableTriple.of(nodeExpressionID,variable,isComposite));
        }
        return variableList;
    }

    public SourceResponse getDisassembly(int sourceReferenceNumber, Runnable done) {

        if( disassembleMap.containsKey(sourceReferenceNumber) ) {
            IDisassembly.IDisassemblyLine[] lines = disassembleMap.get(sourceReferenceNumber);
            StringBuilder sb = new StringBuilder();
            for( IDisassembly.IDisassemblyLine line : lines){
                sb.append(new BigInteger(String.valueOf(line.getAddress())).toString(16) + ":\t");
                sb.append(line.getInstruction()[0].get("Text"));
                sb.append("\n");
            }
            SourceResponse resp = new SourceResponse();
            resp.setContent(sb.toString());
            return resp;
        }
        SourceResponse resp = new SourceResponse();
        return resp;
    }


    public static class ExpressionManager {

        List<String> expressions = new ArrayList<>();

        public void addExpression(String expr ){
            expressions.add(expr);
        }
        public List<String> getExpressions(){
            return expressions;
        }
    }


    private class CreateNodeRunnable implements Runnable {

        final String id;
        final ArrayList<Runnable> waiting_list = new ArrayList<Runnable>();
        final ArrayList<IService> service_list = new ArrayList<IService>();

        CreateNodeRunnable(String id) {
            this.id = id;
            assert context_map.get(id) == null;
            String[] arr = { IRunControl.NAME, IStackTrace.NAME, IRegisters.NAME };
            for (String nm : arr) {
                IService s = channel.getRemoteService(nm);
                if (s != null) service_list.add(s);
            }
            context_map.put(id, this);
            Protocol.invokeLater(this);
        }

        void wait(Runnable r) {
            assert context_map.get(id) == this;
            if (waiting_list.contains(r)) return;
            waiting_list.add(r);
        }

        void done(Object res) {
            assert res != null;
            assert res != this;
            context_map.put(id, res);
            for (Runnable r : waiting_list) {
                if (createNode(id, r)) {
                    if (r instanceof TCFDataCache<?>) ((TCFDataCache<?>)r).post();
                    else r.run();
                }
            }
        }

        public void run() {
            if (context_map.get(id) != this) {
                Object res = context_map.get(id);
                if (res == null) res = new Exception("Invalid context ID");
                done(res);
            }
            else if (service_list.size() == 0) {
                done(new Exception("Invalid context ID"));
            }
            else {
                IService s = service_list.remove(0);
                if (s instanceof IRunControl) {
                    ((IRunControl)s).getContext(id, new IRunControl.DoneGetContext() {
                        public void doneGetContext(IToken token, Exception error, IRunControl.RunControlContext context) {
                            if (error == null && context != null) done(context);
                            else run();
                        }
                    });
                }
                else if (s instanceof IStackTrace) {
                    ((IStackTrace)s).getContext(new String[]{ id }, new IStackTrace.DoneGetContext() {
                        public void doneGetContext(IToken token, Exception error, IStackTrace.StackTraceContext[] context) {
                            if (error == null && context != null && context.length == 1 && context[0] != null) done(context[0]);
                            else run();
                        }
                    });
                }
                else {
                    ((IRegisters)s).getContext(id, new IRegisters.DoneGetContext() {
                        public void doneGetContext(IToken token, Exception error, IRegisters.RegistersContext context) {
                            if (error == null && context != null) done(context);
                            else run();
                        }
                    });
                }
            }
        }
    }


    /**
     * Asynchronously create model node for given ID.
     * If 'done' is TCFDataCache and it is valid after the method returns,
     * the node cannot be created because of an error,
     * and the cache will contain the error report.
     * @param id - context ID.
     * @param done - an object waiting for cache validation.
     * @return - true if all done, false if 'done' is waiting for remote data.
     */
    public boolean createNode(String id, final Runnable done) {
        TCFNode parent = getNode(id);
        if (parent != null) return true;
        LinkedList<Object> path = null;
        for (;;) {
            Object obj = context_map.get(id);
            if (obj == null) obj = new CreateNodeRunnable(id);
            if (obj instanceof CreateNodeRunnable) {
                ((CreateNodeRunnable)obj).wait(done);
                return false;
            }
            if (obj instanceof Throwable) {
                if (done instanceof TCFDataCache<?>) {
                    ((TCFDataCache<?>)done).set(null, (Throwable)obj, null);
                }
                return true;
            }
            if (path == null) path = new LinkedList<Object>();
            path.add(obj);
            String parent_id = null;
            if (obj instanceof IRunControl.RunControlContext) {
                parent_id = ((IRunControl.RunControlContext)obj).getParentID();
            }
            else if (obj instanceof IStackTrace.StackTraceContext) {
                parent_id = ((IStackTrace.StackTraceContext)obj).getParentID();
            }
            else {
                parent_id = ((IRegisters.RegistersContext)obj).getParentID();
            }
            parent = parent_id == null ? launch_node : getNode(parent_id);
            if (parent != null) break;
            id = parent_id;
        }
        while (path.size() > 0) {
            Object obj = path.removeLast();
            if (obj instanceof IRunControl.RunControlContext) {
                IRunControl.RunControlContext ctx = (IRunControl.RunControlContext)obj;
                TCFNodeExecContext n = new TCFNodeExecContext(parent, ctx.getID());
                if (parent instanceof TCFNodeLaunch) ((TCFNodeLaunch)parent).getChildren().add(n);
                else ((TCFNodeExecContext)parent).getChildren().add(n);
                n.setRunContext(ctx);
                parent = n;
            }
            else if (obj instanceof IStackTrace.StackTraceContext) {
                IStackTrace.StackTraceContext ctx = (IStackTrace.StackTraceContext)obj;
                TCFNodeStackFrame n = new TCFNodeStackFrame((TCFNodeExecContext)parent, ctx.getID(), false);
                ((TCFNodeExecContext)parent).getStackTrace().add(n);
                parent = n;
            }
            //TODO:
//            else if (obj instanceof IRegisters.RegistersContext) {
//                IRegisters.RegistersContext ctx = (IRegisters.RegistersContext)obj;
//                TCFNodeRegister n = new TCFNodeRegister(parent, ctx.getID());
//                if (parent instanceof TCFNodeRegister) ((TCFNodeRegister)parent).getChildren().add(n);
//                else if (parent instanceof TCFNodeStackFrame) ((TCFNodeStackFrame)parent).getRegisters().add(n);
//                else ((TCFNodeExecContext)parent).getRegisters().add(n);
//                parent = n;
//            }
            else {
                assert false;
            }
        }
        return true;
    }

    void removeNode(String id) {
        assert id != null;
        assert Protocol.isDispatchThread();
        // Handling of MemoryBlocks.
        //TODO:
       //  for (TCFMemoryBlock b : mem_blocks) b.onContextExited(id);
        id2node.remove(id);
    }



    /**
     * @return communication channel that this model is using.
     */
    public IChannel getChannel() {
        return channel;
    }


    /**
     * Set current hover expression for a given model node,
     * and return a cache of expression nodes that represents given expression.
     * The model allows only one current hover expression per node at any time,
     * however it will cache results of recent expression evaluations,
     * and it will re-use cached results when current hover expression changes.
     * The cache getData() method should not return more then 1 node,
     * and it can return an empty collection.
     * @param parent - a thread or stack frame where the expression should be evaluated.
     * @param expression - the expression text, can be null.
     * @return a cache of expression nodes.
     */
    public TCFChildren getHoverExpressionCache(TCFNode parent, String expression) {
        assert Protocol.isDispatchThread();
        if (parent instanceof TCFNodeStackFrame) {
            return ((TCFNodeStackFrame)parent).getHoverExpressionCache(expression);
        }
        if (parent instanceof TCFNodeExecContext) {
            return ((TCFNodeExecContext)parent).getHoverExpressionCache(expression);
        }
        return null;
    }


    /**
     * Get a model node with given ID.
     * ID == "" means launch node.
     * @param id - node ID.
     * @return debug model node or null if no node exists with such ID.
     */
    public TCFNode getNode(String id) {
        if (id == null) return null;
        if (id.equals("")) return launch_node;
        assert Protocol.isDispatchThread();
        return id2node.get(id);
    }
    /**
     * Get top level (root) debug model node.
     * Same as getNode("").
     * @return root node.
     */
    public TCFNodeLaunch getRootNode() {
        return launch_node;
    }


    private final TCFLaunch.ActionsListener actions_listener = new TCFLaunch.ActionsListener() {

        public void onContextActionStart(TCFAction action) {
            final String id = action.getContextID();
            active_actions.put(id, action);
            // updateAnnotations(null); // Eclipse Annotations in debug window
        }
        public void onContextActionResult(String id, String reason) {
            if (reason == null) action_results.remove(id);
            else action_results.put(id, reason);
        }
        public void onContextActionDone(final TCFAction action) {
            Runnable r = new Runnable() {
                @Override
                public void run() {
                    TCFContextState state_data = null;
                    String id = action.getContextID();
                    TCFNode node = getNode(id);
                    if (node instanceof TCFNodeExecContext) {
                        TCFDataCache<TCFContextState> state_cache = ((TCFNodeExecContext)node).getState();
                        if (!state_cache.validate(this)) return;
                        state_data = state_cache.getData();
                    }
                    if (active_actions.get(id) == action) {
                        active_actions.remove(id);
                        if (node instanceof TCFNodeExecContext) {
                            ((TCFNodeExecContext)node).onContextActionDone();
                            if (state_data != null && state_data.is_suspended) {
                                String reason = action_results.get(id);
                                if (reason == null) reason = state_data.suspend_reason;
                                // setDebugViewSelection(id2node.get(id), reason);
                            }
                        }
//                        for (TCFModelProxy p : model_proxies) p.post();
//                        updateAnnotations(null);
//                        TCFNodePropertySource.refresh(node);
                    }
                    launch.removePendingClient(this);
                }
            };
            launch.addPendingClient(r);
            Protocol.invokeLater(r);
        }
    };

    private final IRunControl.RunControlListener run_listener = new IRunControl.RunControlListenerV1() {

        // For Linux usecase, conainer context callbacks will not be called.
        public void containerResumed(String[] context_ids) {
            for (String id : context_ids) {
                TCFNode node = getNode(id);
                if (node instanceof TCFNodeExecContext) {
                    ((TCFNodeExecContext)node).onContainerResumed();
                }
            }
            // How do we inform IDE about this?
            //  updateAnnotations(null);
        }

        public void containerSuspended(String context, String pc, String reason,
                                       Map<String,Object> params, String[] suspended_ids) {
            boolean func_call = false;
            if (params != null) {
                Boolean b = (Boolean)params.get(IRunControl.STATE_FUNC_CALL);
                func_call = b != null && b.booleanValue();
            }
            int action_cnt = 0;
            for (String id : suspended_ids) {
                TCFNode node = getNode(id);
                action_results.remove(id);
                if (active_actions.get(id) != null) action_cnt++;
                if (!id.equals(context) && node instanceof TCFNodeExecContext) {
                    ((TCFNodeExecContext)node).onContainerSuspended(func_call);
                }
                // onMemoryChanged(id, true, true, false);
            }
            TCFNode node = getNode(context);
            if (node instanceof TCFNodeExecContext) {
                ((TCFNodeExecContext)node).onContextSuspended(pc, reason, params, func_call);
            }
            launch_node.onAnyContextSuspendedOrChanged();
            //TODO: commenting below block code.
//            if (!func_call && action_cnt == 0) {
//                setDebugViewSelection(node, reason);
//                updateAnnotations(null);
//                TCFNodePropertySource.refresh(node);
//            }
            action_results.remove(context);
        }

        public void contextAdded(IRunControl.RunControlContext[] contexts) {
            for (IRunControl.RunControlContext ctx : contexts) {
                String id = ctx.getParentID();
                if (id == null) {
                    launch_node.onContextAdded(ctx);
                }
                else {
                    TCFNode node = getNode(id);
                    if (node instanceof TCFNodeExecContext) {
                        ((TCFNodeExecContext)node).onContextAdded(ctx);
                    }
                }
                context_map.put(ctx.getID(), ctx);
            }
            launch_node.onAnyContextAddedOrRemoved();
        }

        public void contextChanged(IRunControl.RunControlContext[] contexts) {
            for (IRunControl.RunControlContext ctx : contexts) {
                String id = ctx.getID();
                context_map.put(id, ctx);
                TCFNode node = getNode(id);
                if (node instanceof TCFNodeExecContext) {
                    ((TCFNodeExecContext)node).onContextChanged(ctx);
                }
                //TODO:
//                onMemoryChanged(id, true, false, false);
//                if (active_actions.get(id) == null) {
//                    TCFNodePropertySource.refresh(node);
//                }
            }
            launch_node.onAnyContextSuspendedOrChanged();
        }

        public void contextException(String context, String msg) {
            TCFNode node = getNode(context);
            if (node instanceof TCFNodeExecContext) {
                ((TCFNodeExecContext)node).onContextException(msg);
            }
        }

        public void contextRemoved(final String[] context_ids) {
            onContextRemoved(context_ids);
        }

        public void contextResumed(String id) {
            TCFNode node = getNode(id);
            if (node instanceof TCFNodeExecContext) {
                ((TCFNodeExecContext)node).onContextResumed();
            }
            //TODO:
            // updateAnnotations(null);
        }

        public void contextSuspended(String id, String pc, String reason, Map<String,Object> params) {
            boolean func_call = false;
            // CHECK_NEXT On suspended.
            if (params != null) {
                Boolean b = (Boolean)params.get(IRunControl.STATE_FUNC_CALL);
                func_call = b != null && b.booleanValue();
            }
            TCFNode node = getNode(id);
            action_results.remove(id);
            if (node instanceof TCFNodeExecContext) {
                ((TCFNodeExecContext)node).onContextSuspended(pc, reason, params, func_call);
            }
            launch_node.onAnyContextSuspendedOrChanged();
            //TODO:
            if (!func_call && active_actions.get(id) == null) {
                setDebugViewSelection(node, reason);
//                updateAnnotations(null);
//                TCFNodePropertySource.refresh(node);
            }
//            onMemoryChanged(id, true, true, false);
        }
        /**
         * Handle debug view selection and suspend triggers when debug context is suspended.
         * @param node - model node that represents the debug context.
         * @param reason - suspend reason.
         */
        public void setDebugViewSelection(TCFNode node, String reason) {
            assert Protocol.isDispatchThread();
            if (node == null) return;
            if (node.isDisposed()) return;
            //launch.clientOut("Reason: " + reason);
//            if (!SELECT_ADDED.equals(reason)) runSuspendTrigger(node);
//            if (initial_selection != null) Protocol.invokeLater(initial_selection);
//            if (reason == null) return;
//            for (TCFModelProxy proxy : model_proxies) {
//                if (proxy.getPresentationContext().getId().equals(IDebugUIConstants.ID_DEBUG_VIEW)) {
//                    setDebugViewSelectionForProxy(proxy, node, reason);
//                }
//                if (reason.equals(IRunControl.REASON_BREAKPOINT)) {
//                    IWorkbenchPart part = proxy.getPresentationContext().getPart();
//                    int policy = getLockPolicy(part);
//                    if (policy == UPDATE_POLICY_BREAKPOINT) {
//                        clearLock(part);
//                        setLock(part);
//                    }
//                }
//            }
        }


        public void contextStateChanged(String id) {
            TCFNode node = getNode(id);
            if (node instanceof TCFNodeExecContext) {
                ((TCFNodeExecContext)node).onContextStateChanged();
            }
        }
    };

    private void onContextRemoved(String[] context_ids) {
        HashSet<String> set = new HashSet<String>();
        for (String id : context_ids) {
            launch_node.onContextRemoved(id);
            TCFNode node = getNode(id);
            if (node instanceof TCFNodeExecContext) {
                ((TCFNodeExecContext)node).onContextRemoved();
                //TODO:
//                for (TCFModelProxy p : model_proxies) {
//                    p.saveExpandState(node);
//                    p.clearAutoExpandStack(id);
//                }
            }
            action_results.remove(id);
            //TODO:
//            if (mem_blocks_update != null) mem_blocks_update.changeset.remove(id);
            set.add(id);
        }

        for (;;) {
            int n = set.size();
            for (Map.Entry<String,Object> e : context_map.entrySet()) {
                Object obj = e.getValue();
                if (obj instanceof IRunControl.RunControlContext) {
                    IRunControl.RunControlContext x = (IRunControl.RunControlContext)obj;
                    if (set.contains(x.getParentID())) set.add(x.getID());
                    String pid = x.getProcessID();
                    if (pid != null && set.contains(pid)) set.add(x.getID());
                }
                else if (obj instanceof IStackTrace.StackTraceContext) {
                    IStackTrace.StackTraceContext x = (IStackTrace.StackTraceContext)obj;
                    if (set.contains(x.getParentID())) set.add(x.getID());
                }
                else if (obj instanceof IRegisters.RegistersContext) {
                    IRegisters.RegistersContext x = (IRegisters.RegistersContext)obj;
                    if (set.contains(x.getParentID())) set.add(x.getID());
                    String pid = x.getProcessID();
                    if (pid != null && set.contains(pid)) set.add(x.getID());
                }
                else if (obj instanceof Throwable) {
                    set.add(e.getKey());
                }
            }
            if (n == set.size()) break;
        }
        context_map.keySet().removeAll(set);

        launch_node.onAnyContextAddedOrRemoved();
        // Close debug session if the last context is removed:
        onContextOrProcessRemoved();
        //TODO:
        // updateAnnotations(null);
    }

    private void onContextOrProcessRemoved() {
        final int generation = ++auto_disconnect_generation;
        Protocol.invokeLater(1000, new Runnable() {
            public void run() {
                if (generation != auto_disconnect_generation) return;
                if (launch_node == null) return;
                if (launch_node.isDisposed()) return;
                TCFChildren children = launch_node.getFilteredChildren();
                if (!children.validate(this)) return;
                if (children.size() > 0) return;
                launch.onLastContextRemoved();
            }
        });
    }

    private final IProcesses.ProcessesListener prs_listener = new IProcesses.ProcessesListener() {

        public void exited(String process_id, int exit_code) {
            IProcesses.ProcessContext prs = launch.getProcessContext();
            if (prs != null && process_id.equals(prs.getID())) onContextOrProcessRemoved();
        }
    };
    private class InitialSuspendTrigger extends InitialRunnable {
        public InitialSuspendTrigger() {
            initial_suspend_trigger = this;
        }
        @Override
        public void run() {
            if (launch_node == null || initial_suspend_trigger != this) return;
            ArrayList<TCFNodeExecContext> nodes = new ArrayList<TCFNodeExecContext>();
            if (!searchSuspendedThreads(launch_node.getFilteredChildren(), nodes)) return;
            if (nodes.size() > 0) runSuspendTrigger(nodes.get(0));
        }
    };
    /*
     * Lazily run registered suspend listeners.
     * @param node - suspended context.
     */
    private synchronized void runSuspendTrigger(final TCFNode node) {
        initial_suspend_trigger = null;
        this.launch.clientStop();
        //launch.clientOut("Hit suspend trigger");
        // TODO:
//        if (suspend_trigger_listeners.size() == 0) return;
//        final ISuspendTriggerListener[] listeners = suspend_trigger_listeners.toArray(
//                new ISuspendTriggerListener[suspend_trigger_listeners.size()]);
//
//        final int generation = ++suspend_trigger_generation;
//        if (wait_for_views_update_after_step) {
//            launch.addPendingClient(suspend_trigger_listeners);
//        }
//        asyncExec(new Runnable() {
//            public void run() {
//                synchronized (TCFModel.this) {
//                    if (generation != suspend_trigger_generation) return;
//                }
//                for (final ISuspendTriggerListener listener : listeners) {
//                    try {
//                        listener.suspended(launch, node);
//                    }
//                    catch (Throwable x) {
//                        Activator.log(x);
//                    }
//                }
//                synchronized (TCFModel.this) {
//                    if (generation != suspend_trigger_generation) return;
//                    launch.removePendingClient(suspend_trigger_listeners);
//                }
//            }
//        });
    }


    private abstract class InitialRunnable implements Runnable {
        protected boolean searchSuspendedThreads(TCFChildren c, ArrayList<TCFNodeExecContext> nodes) {
            if (!c.validate(this)) return false;
            for (TCFNode n : c.toArray()) {
                if (!searchSuspendedThreads((TCFNodeExecContext)n, nodes)) return false;
            }
            return true;
        }
        private boolean searchSuspendedThreads(TCFNodeExecContext n, ArrayList<TCFNodeExecContext> nodes) {
            TCFDataCache<IRunControl.RunControlContext> run_context = n.getRunContext();
            if (!run_context.validate(this)) return false;
            IRunControl.RunControlContext ctx = run_context.getData();
            if (ctx != null && ctx.hasState()) {
                TCFDataCache<TCFContextState> state = n.getState();
                if (!state.validate(this)) return false;
                TCFContextState s = state.getData();
                if (s != null && s.is_suspended) nodes.add(n);
                return true;
            }
            return searchSuspendedThreads(n.getChildren(), nodes);
        }
    }


    void onConnected() {
        assert Protocol.isDispatchThread();
        assert launch_node == null;
        channel = launch.getChannel();
        launch_node = new TCFNodeLaunch(this);
        //TODO:
//        IMemory mem = launch.getService(IMemory.class);
        // if (mem != null) mem.addListener(mem_listener);
        IRunControl run = launch.getService(IRunControl.class);
        if (run != null) run.addListener(run_listener);
        // TODO:
//        IMemoryMap mmap = launch.getService(IMemoryMap.class);
//        if (mmap != null) mmap.addListener(mmap_listener);
//        IPathMap pmap = launch.getService(IPathMap.class);
//        if (pmap != null) pmap.addListener(pmap_listener);
//        IRegisters reg = launch.getService(IRegisters.class);
//        if (reg != null) reg.addListener(reg_listener);
        IProcesses prs = launch.getService(IProcesses.class);
        if (prs != null) prs.addListener(prs_listener);
        //TODO: Used for changing the launch view.
        // launchChanged();
        Protocol.invokeLater(new InitialSuspendTrigger());
        // TODO: TCFModelProxy is about debug views.
//        for (TCFModelProxy p : model_proxies) {
//            String id = p.getPresentationContext().getId();
//            if (IDebugUIConstants.ID_DEBUG_VIEW.equals(id)) {
//                Protocol.invokeLater(new InitialSelection(p));
//            }
//        }

        if (launch.isProcessExited()) onContextOrProcessRemoved();
        // TODO:
        // for (TCFConsole c : process_consoles.values()) c.onModelConnected();
    }

    public boolean getStackFramesLimitEnabled() {
        return stack_frames_limit_enabled;
    }

    public int getStackFramesLimitValue() {
        return stack_frames_limit_value;
    }

    void onDisconnected() {
        assert Protocol.isDispatchThread();
        // TODO:
//        if (locks.size() > 0) {
//            TCFSnapshot[] arr = locks.values().toArray(new TCFSnapshot[locks.size()]);
//            locks.clear();
//            for (TCFSnapshot s : arr) s.dispose();
//        }
        if (launch.getChannel() != null) {
            // TODO:
//            IMemory mem = launch.getService(IMemory.class);
//            if (mem != null) mem.removeListener(mem_listener);
            IRunControl run = launch.getService(IRunControl.class);
            if (run != null) run.removeListener(run_listener);
            // TODO:
//            IMemoryMap mmap = launch.getService(IMemoryMap.class);
//            if (mmap != null) mmap.removeListener(mmap_listener);
//            IPathMap pmap = launch.getService(IPathMap.class);
//            if (pmap != null) pmap.removeListener(pmap_listener);
//            IRegisters reg = launch.getService(IRegisters.class);
//            if (reg != null) reg.removeListener(reg_listener);
            IProcesses prs = launch.getService(IProcesses.class);
            if (prs != null) prs.removeListener(prs_listener);
        }
        if (launch_node != null) {
            launch_node.dispose();
            launch_node = null;
        }
        // Dispose memory monitors
        // TODO:
//        TCFMemoryBlock.onModelDisconnected(this);
//        mem_retrieval.clear();
//        mem_blocks.clear();
        // Refresh the Debug view - cannot be done through ModelProxy since it is disposed
        // TODO:
        // refreshLaunchView();
        assert id2node.size() == 0;
    }
    //TODO: Pushing the data to IDE, instead of console.
    void onProcessOutput(String ctx_id, int stream_id, byte[] data) {
        if(data != null)
            launch.clientOut(new String(data));
    }

    void onProcessStreamError(String process_id, int stream_id, Exception x, int lost_size) {
        if (channel == null || channel.getState() == IChannel.STATE_CLOSED) return;
        StringBuffer bf = new StringBuffer();
        bf.append("Debugger console IO error");
        if (process_id != null) {
            bf.append(". Process ID ");
            bf.append(process_id);
        }
        bf.append(". Stream ");
        bf.append(stream_id);
        if (lost_size > 0) {
            bf.append(". Lost data size ");
            bf.append(lost_size);
        }
        launch.clientErr(bf.toString() + x.toString());
    }
    String getContextActionResult(String id) {
        return action_results.get(id);
    }
    private volatile boolean instruction_stepping_enabled;
    /**
     * @return whether instruction stepping is enabled
     */
    public boolean isInstructionSteppingEnabled() {
        return instruction_stepping_enabled;
    }

    public void getMatchingProcessRunContext(final String processId, CompletableFuture<TCFDataCache<IRunControl.RunControlContext>> runContextFuture) {
        Protocol.invokeLater(25, new Runnable() {
            public void run() {

                TCFNodeLaunch node = (TCFNodeLaunch) getNode("");
                TCFChildren children = node.getChildren();
                Map<String, ITCFObject> children1 = children.getChildren();
                children1.entrySet().stream().forEach(e -> {
                    String key = e.getKey();
                    if (!key.equalsIgnoreCase(processId)) return;
                    TCFNodeExecContext tcfNodeExecContext = (TCFNodeExecContext) e.getValue();
                    TCFDataCache<IRunControl.RunControlContext> runContext = tcfNodeExecContext.getRunContext();
                    if (!runContext.validate(this)) return;
                    if (!runContext.isDisposed()) {
                        runContextFuture.complete(runContext);
                    } else {
                        runContextFuture.complete(null);
                    }
                });
            }
        });
    }
    /**
     * @return whether to filter variant fields by discriminant value
     */
    public boolean isFilterVariantsByDiscriminant() {
        return filter_variants_by_discriminant;
    }
    /**
     * @return whether to show qualified type names.
     */
    public boolean isShowQualifiedTypeNamesEnabled() {
        return qualified_type_names_enabled;
    }

    /**
     * Search memory context that owns the object represented by given node.
     * @return data cache item that holds the memory context node.
     */
    public TCFDataCache<TCFNodeExecContext> searchMemoryContext(final TCFNode node) {
        TCFNode n = node;
        while (n != null && !n.isDisposed()) {
            if (n instanceof TCFNodeExecContext) return ((TCFNodeExecContext)n).getMemoryNode();
            n = n.parent;
        }
        return null;
    }



    /**
     * Get a type that should be used to cast a value of an expression when it is shown in a view.
     * Return null if original type of the value should be used.
     * @param id - expression node ID.
     * @return a string that designates a type or null.
     */
    public String getCastToType(String id) {
        return cast_to_type_map.get(id);
    }

    /**
     * Register a type that should be used to cast a value of an expression when it is shown in a view.
     * 'type' == null means original type of the value should be used.
     * @param id - expression node ID.
     * @param type - a string that designates a type.
     */
    public void setCastToType(String id, String type) {
        if (type != null && type.trim().length() == 0) type = null;
        if (type == null) cast_to_type_map.remove(id);
        else cast_to_type_map.put(id, type);
        TCFNode node = id2node.get(id);
        if (node instanceof ICastToType) {
            ((ICastToType)node).onCastToTypeChanged();
        }
    }

    /**
     * Get a data cache that contains location info of a symbol.
     * New cache object is created if it does not exist yet.
     * @param sym_id - the symbol ID.
     * @return data cache object.
     */
    public TCFDataCache<Map<String,Object>> getSymbolLocationCache(final String sym_id) {
        if (sym_id == null) return null;
        TCFNodeSymbol n = (TCFNodeSymbol)getNode(sym_id);
        if (n == null) n = new TCFNodeSymbol(launch_node, sym_id);
        return n.getLocation();
    }

    /**
     * Create human readable error message from a Throwable object.
     * @param error - a Throwable object.
     * @param multiline - true if multi-line text is allowed.
     * @return error message text.
     */
    public static String getErrorMessage(Throwable error, boolean multiline) {
        StringBuffer buf = new StringBuffer();
        while (error != null) {
            String msg = null;
            if (!multiline && error instanceof IErrorReport) {
                msg = Command.toErrorString(((IErrorReport)error).getAttributes());
            }
            else if (error instanceof UnknownHostException) {
                msg = "Unknown host: " + error.getMessage();
            }
            else {
                msg = error.getLocalizedMessage();
            }
            if (msg == null || msg.length() == 0) msg = error.getClass().getName();
            buf.append(msg);
            error = error.getCause();
            if (error != null) {
                char ch = buf.charAt(buf.length() - 1);
                if (multiline && ch != '\n') {
                    buf.append('\n');
                }
                else if (ch != '.' && ch != ';') {
                    buf.append(';');
                }
                if (multiline) buf.append("Caused by:\n");
                else buf.append(' ');
            }
        }
        if (buf.length() > 0) {
            char ch = buf.charAt(buf.length() - 1);
            if (multiline && ch != '\n') {
                buf.append('\n');
            }
        }
        return buf.toString();
    }


    /**
     * Get a data cache that contains children of a symbol.
     * New cache object is created if it does not exist yet.
     * @param sym_id - the symbol ID.
     * @return data cache object.
     */
    public TCFDataCache<String[]> getSymbolChildrenCache(final String sym_id) {
        if (sym_id == null) return null;
        TCFNodeSymbol n = (TCFNodeSymbol)getNode(sym_id);
        if (n == null) n = new TCFNodeSymbol(launch_node, sym_id);
        return n.getChildren();
    }
    private static class AddressRange {
        BigInteger start;
        BigInteger end;
    }

    public  ImmutablePair<StackFrame, String>  displaySource(final Object element, Runnable done) {
        // final Object generation = displaySourceStart(page, wait_for_pc_update_after_step);
        assert Protocol.isDispatchThread();
                // if (!displaySourceCheck(page, generation)) return;
                TCFNodeStackFrame stack_frame = null;
        final TCFNodeExecContext[] exec_ctx = {null};
                if (!disposed && channel.getState() == IChannel.STATE_OPEN) {
                    if (element instanceof TCFNodeExecContext) {
                        exec_ctx[0] = (TCFNodeExecContext)element;
                        if (!exec_ctx[0].isDisposed() && active_actions.get(exec_ctx[0].id) == null) {
                            TCFDataCache<TCFContextState> state_cache = exec_ctx[0].getState();
                            if (!state_cache.validate(done)) return null;
                            if (!exec_ctx[0].isNotActive()) {
                                TCFContextState state_data = state_cache.getData();
                                if (state_data != null && state_data.is_suspended) {
                                    TCFChildrenStackTrace stack_trace = exec_ctx[0].getStackTrace();
                                    if (!stack_trace.validate(done)) return null;
                                    stack_frame = stack_trace.getTopFrame();
                                }
                            }
                        }
                    }
                    else if (element instanceof TCFNodeStackFrame) {
                        TCFNodeStackFrame f = (TCFNodeStackFrame)element;
                        exec_ctx[0] = (TCFNodeExecContext)f.parent;
                        if (!f.isDisposed() && !exec_ctx[0].isDisposed() && active_actions.get(exec_ctx[0].id) == null) {
                            TCFDataCache<TCFContextState> state_cache = exec_ctx[0].getState();
                            if (!state_cache.validate(done)) return null;
                            if (!exec_ctx[0].isNotActive()) {
                                TCFContextState state_data = state_cache.getData();
                                if (state_data != null && state_data.is_suspended) {
                                    // Validate stack trace to make sure stack_frame.getFrameNo() is valid
                                    TCFChildrenStackTrace stack_trace = exec_ctx[0].getStackTrace();
                                    if (!stack_trace.validate(done)) return null;
                                    stack_frame = f;
                                }
                            }
                        }
                    }
                }
                String ctx_id = null;
                String mem_id = null;
                boolean top_frame = false;
                ILineNumbers.CodeArea area = null;
                if (stack_frame != null) {
                    TCFDataCache<TCFSourceRef> line_info = stack_frame.getLineInfo();
                    if (!line_info.validate(done)) return null;
                    Throwable error = line_info.getError();
                    TCFSourceRef src_ref = line_info.getData();
                    if (error == null && src_ref != null) error = src_ref.error;
                    if (error != null) {
                        LOGGER.fine("Error retrieving source mapping for a stack frame"+ error.toString());
                    }
                    if (src_ref != null) {
                        mem_id = src_ref.context_id;
                        area = src_ref.area;
                    }
                    top_frame = stack_frame.getFrameNo() == 0;
                    ctx_id = stack_frame.parent.id;

                }
                StackFrame stackFrame = new StackFrame();
                if(area != null ) {
                    stackFrame.setLine(area.start_line);
                    // stackFrame.setEndLine(area.end_line);
                    stackFrame.setColumn(area.start_column);
                    // stackFrame.setEndColumn(area.end_column);
                    Source source = new Source();
                    source.setPath(TCFSourceLookupDirector.lookup(launch, area));
                    source.setName(area.file);
                    source.setPresentationHint(SourcePresentationHint.EMPHASIZE);
                    stackFrame.setSource(source);
                } else if(stack_frame != null ) {

                    TCFNodeStackFrame finalStack_frame = stack_frame;

                    /*
                    Protocol.invokeLater(new Runnable() {
                                             IMemory.MemoryContext mem;
                                             boolean big_endian;
                                             int addr_bits;
                                             IDisassembly.IDisassemblyLine[] disassembly;
                                             AddressRange range;
                                             boolean done_disassembly;
                                             ISymbols.Symbol[] symbol_array;
                                             boolean done_symbols;
                                             ILineNumbers.CodeArea[] code_areas;
                                             boolean done_line_numbers;
                                             byte[] code;
                                             boolean done_code;

                                                 @Override
                                                 public void run() {
                                                     TCFNodeStackFrame frame = finalStack_frame;
                                                     TCFDataCache<BigInteger> addr = frame.getAddress();
                                                     if (!addr.validate(this)) return;
                                                     BigInteger startAddress = addr.getData();
                                                     BigInteger endAddr = startAddress.add(BigInteger.valueOf(128));
                                                     int lines = 32;
                                                     int linesHint = 32;
                                                     TCFNode parent = exec_ctx[0].getParent();
                                                     if(parent instanceof TCFNodeExecContext ){
                                                         exec_ctx[0] = (TCFNodeExecContext) exec_ctx[0].getParent();
                                                     }

                                                     // retrieveDisassembly(pcAddr,endAddr,lines,exec_ctx);
                                                     IDisassembly disass = channel.getRemoteService(IDisassembly.class);
                                                     TCFDataCache<IMemory.MemoryContext> cache = exec_ctx[0].getMemoryContext();
                                                     if (!cache.validate(this)) return;

                                                     mem = cache.getData();
                                                     if (mem != null) {
                                                         big_endian = mem.isBigEndian();
                                                         addr_bits = mem.getAddressSize() * 8;
                                                         int accessSize = 0;
                                                         BigInteger mem_end = BigInteger.ONE.shiftLeft(addr_bits);
                                                         mem_end = mem_end.subtract(BigInteger.ONE);
                                                         final BigInteger requestedLineEndAddr = startAddress.add(BigInteger.valueOf(linesHint * mem.getAddressSize()));

                                                         if (requestedLineEndAddr.compareTo(mem_end) > 0) {
                                                             accessSize = mem_end.subtract(startAddress).intValue() + 1;
                                                         }
                                                         else {
                                                             accessSize = linesHint * mem.getAddressSize();
                                                         }


                                                         if(! done_disassembly ) {
                                                             Map<String, Object> params = new HashMap<String, Object>();
                                                             disass.disassemble(mem.getID(), startAddress, accessSize, params, new IDisassembly.DoneDisassemble() {
                                                                 @Override
                                                                 public void doneDisassemble(IToken token, final Throwable error, IDisassembly.IDisassemblyLine[] res) {
//                                    if (error != null) {
//                                        fCallback.asyncExec(new Runnable() {
//                                            public void run() {
//                                                insertError(request, startAddress, error);
//                                                if (fCallback.getAddressSize() < addr_bits) fCallback.addressSizeChanged(addr_bits);
//                                            }
//                                        });
//                                        return;
//                                    }
                                                                     if (res != null && res.length > 0) {
                                                                         disassembly = res;
                                                                         range = new AddressRange();
                                                                         range.start = JSON.toBigInteger(res[0].getAddress());
                                                                         IDisassembly.IDisassemblyLine last = res[res.length - 1];
                                                                         range.end = JSON.toBigInteger(last.getAddress()).add(BigInteger.valueOf(last.getSize()));
                                                                     }
                                                                     done_disassembly = true;
                                                                     run();
                                                                 }
                                                             });
                                                             return;
                                                         }
                                                         IDisassembly.IDisassemblyLine line = disassembly[0];
                                                     }
                                                 }
                                             }
                        );

                     */
                    TCFNode parent = exec_ctx[0].getParent();
                    if( parent instanceof TCFNodeExecContext ){
                        TCFDataCache<BigInteger> addr = finalStack_frame.getAddress();
                        if (!addr.validate(done )) return null;
                        BigInteger data = addr.getData();
                        int line_number = 1;
                        if(data != null ) {
                            String addrStr = data.toString(16);
                            ImmutablePair<Integer,Integer> sourceRefLinePair = checkAndGetDisassemblyLineNumber(data,finalStack_frame,done);
                            if(sourceRefLinePair == null ) return null;
                            line_number = sourceRefLinePair.right;
                            Source source = new Source();
                            stackFrame.setLine(line_number);
                            stackFrame.setColumn(1);
                            source.setName("Disassembly");
                            source.setSourceReference(sourceRefLinePair.left);
                            stackFrame.setSource(source);
                        }
                    }
                }

                if(stack_frame != null && ( !stack_frame.getID().equalsIgnoreCase("0" ) ) && (stack_frame.getFrameNo() != -1 ) ) {
                    // CompletableFuture<String> labelFuture = new CompletableFuture<>();
                    TCFDataCache<String> labelInfo = stack_frame.getLabelInfo();
                    if(!labelInfo.validate(done)) return null;
                    // Test code to check the variables.
//                    TCFChildren var_children = stack_frame.getChildren();
//                    if(!var_children.validate(this)) return;
//
//                    Map<String, ITCFObject> children = var_children.getChildren();
//                    for( Map.Entry<String,ITCFObject> e: children.entrySet()){
//                        ITCFObject value = e.getValue();
//                        TCFNodeExpression nodeExpression = (TCFNodeExpression) value;
//
//                        TCFDataCache<String> expressionLabel = nodeExpression.getLabelInfo();
//                        if( !expressionLabel.validate(this)) return;
//                        String expressionLabelData = expressionLabel.getData();
//                    }
                    String label = labelInfo.getData();
                    stackFrame.setName(label);
                    int sfUniqueId = toStackFrameUniqueId(stack_frame.getParent().getID(), stack_frame.getID());
                    stackFrame.setId(sfUniqueId);
                    return ImmutablePair.of(stackFrame,stack_frame.getID());
                }

                return ImmutablePair.of(null,"");
                // displaySource(element, ctx_id, mem_id, top_frame, area);

    }

    private ImmutablePair<Integer,Integer> checkAndGetDisassemblyLineNumber(BigInteger data, TCFNodeStackFrame finalStack_frame, Runnable done) {
        for (Map.Entry<Integer, IDisassembly.IDisassemblyLine[]> entry : disassembleMap.entrySet()) {
            Optional<Integer> lineNumberOptional = getMatchingDisassemblyLine(data, entry.getValue());
            if(lineNumberOptional.isPresent()) {
                return ImmutablePair.of(entry.getKey(),lineNumberOptional.get());
            }
        }
        if(!updateDisassembly(finalStack_frame,done))
            return null;
        for (Map.Entry<Integer, IDisassembly.IDisassemblyLine[]> entry : disassembleMap.entrySet()) {
            Optional<Integer> lineNumberOptional = getMatchingDisassemblyLine(data, entry.getValue());
            if(lineNumberOptional.isPresent()) {
                return ImmutablePair.of(entry.getKey(),lineNumberOptional.get());
            }
        }
        return ImmutablePair.of(0,0);
    }
    private Boolean updateDisassembly(TCFNodeStackFrame nodeStackFrame, Runnable done){

         TCFNode.TCFData<IDisassembly.IDisassemblyLine[]> disassembly = nodeStackFrame.getDisassembly();
         if(!disassembly.validate(done)) return false;
         IDisassembly.IDisassemblyLine[] lines = disassembly.getData();
         StringBuilder sb = new StringBuilder();
         for( IDisassembly.IDisassemblyLine line : lines ){
             sb.append(new BigInteger(String.valueOf(line.getAddress())).toString(16) + ":\t");
             sb.append(line.getInstruction()[0].get("Text"));
             sb.append("\n");
         }
         disassembleMap.put(Math.abs(String.valueOf(lines[0].getAddress()).hashCode()),lines);
         return true;

    }

    private Optional<Integer> getMatchingDisassemblyLine(BigInteger data, IDisassembly.IDisassemblyLine[] disassemblyLines) {
        int disSize = disassemblyLines.length;
        BigInteger startAddress = new BigInteger(String.valueOf(disassemblyLines[0].getAddress()));
        BigInteger endAddress = new BigInteger(String.valueOf(disassemblyLines[disSize - 1].getAddress()));
        int line_number = 1;
        if (data.compareTo(startAddress) >= 0 && data.compareTo(endAddress) < 0) {
            for (int i = 0; i < disassemblyLines.length; i++) {
                if (0 == data.compareTo(new BigInteger(String.valueOf(disassemblyLines[i].getAddress())))) {
                    line_number = i + 1;
                    return Optional.of(line_number);
                }
                if( i+1 <disSize && ( ( data.compareTo(new BigInteger(String.valueOf(disassemblyLines[i].getAddress()))) >0 ) &&
                        ( data.compareTo(new BigInteger(String.valueOf(disassemblyLines[i+1].getAddress()))) < 0 ) )){
                    line_number = i + 1;
                    return Optional.of(line_number);
                }
            }
            return Optional.empty();
        }
        return Optional.empty();
    }

    private void retrieveDisassembly(BigInteger pcAddr, BigInteger endAddr, int lines, TCFNodeExecContext exec_ctx) {
        IDisassembly disass = channel.getRemoteService(IDisassembly.class);
        exec_ctx.getMemoryContext();

    }

    private int toStackFrameUniqueId(String threadContextId, String frameNumber){
        threadContextId = threadContextId.replace("P","");
        threadContextId = threadContextId.replace(".","");
        String fpNum = frameNumber.split("\\.")[0].replace("FP", "");
        return Integer.parseInt(threadContextId + fpNum);
    }
    public TCFAction getActiveAction(String id) {
        return active_actions.get(id);
    }

    private void displaySource(Object element, String ctx_id, String mem_id, boolean top_frame, ILineNumbers.CodeArea area) {
        LOGGER.info(" Got display Source");
    }
    public String  getLabel(Object node, Runnable runnable){
        assert !(node instanceof TCFLaunch);
        return (((TCFNode)node).getLabel(runnable));
    }
}

