/*******************************************************************************
 * Copyright (c) 2022  Xilinx, Inc. All rights reserved.
 * All rights reserved. This software is made available under
 * the terms of the Eclipse Public License 2.0
 * which accompanies this distribution, and is available at
 * https://www.eclipse.org/legal/epl-2.0/
 *
 * Xilinx, Inc.
 * Copyright 2022  Xilinx, Inc. All Rights Reserved.
 *******************************************************************************/

package org.eclipse.tcf.internal.dap.breakpoint;

import java.util.Objects;

public class BreakpointInfo {

    public enum Type {
        SOURCE, FUNCTION, DATA, EXCEPTION;
    }

    private int id;
    private Type type;
    private boolean foreign;
    private boolean enabled = true;
    private String file;
    private Integer line;
    private Integer column;
    private String function;
    private String condition;
    private int ignoreCount;
    private String logMessage;
    private String bpIdString;

    public BreakpointInfo() {
    }

    public static BreakpointInfo newLocalBreakpoint(Type type, int id) {
        return new BreakpointInfo().setType(type).setId(id).setEnabled(true);
    }

    public static BreakpointInfo newForeignBreakpoint(String foreignId, int id) {
        BreakpointInfo info = new BreakpointInfo().setForeign(true).setId(id).setEnabled(true);
        info.bpIdString = foreignId;
        return info;
    }

    public String getBreakpointId() {
        if (isForeign()) {
            return bpIdString;
        }
        else if (bpIdString == null) {
            StringBuilder idBuilder = new StringBuilder();
            if (type == BreakpointInfo.Type.FUNCTION) {
                idBuilder.append(String.format("%s()", function));
            } else if (type == Type.SOURCE) {
                idBuilder.append(file);
                idBuilder.append("[");
                idBuilder.append(line);
                if (column != null) {
                    idBuilder.append(",");
                    idBuilder.append(column);
                }
                idBuilder.append("]");
            }
            idBuilder.append("@");
            idBuilder.append(id);
            bpIdString = idBuilder.toString();
        }
        return bpIdString;
    }

    public int getId() {
        return id;
    }

    public BreakpointInfo setId(int id) {
        this.id = id;
        return this;
    }

    public Type getType() {
        return type;
    }

    public BreakpointInfo setType(Type type) {
        this.type = type;
        return this;
    }

    public boolean isEnabled() {
        return enabled;
    }

    public BreakpointInfo setEnabled(boolean enabled) {
        this.enabled = enabled;
        return this;
    }

    public String getFile() {
        return file;
    }

    public BreakpointInfo setFile(String file) {
        this.file = file;
        return this;
    }

    public Integer getLine() {
        return line;
    }

    public BreakpointInfo setLine(Integer line) {
        this.line = line;
        return this;
    }

    public Integer getColumn() {
        return column;
    }

    public BreakpointInfo setColumn(Integer column) {
        this.column = column;
        return this;
    }

    public String getFunction() {
        return function;
    }

    public BreakpointInfo setFunction(String function) {
        this.function = function;
        return this;
    }

    public String getCondition() {
        return condition;
    }

    public BreakpointInfo setCondition(String condition) {
        this.condition = condition;
        return this;
    }

    public int getIgnoreCount() {
        return ignoreCount;
    }

    public BreakpointInfo setIgnoreCount(int ignoreCount) {
        this.ignoreCount = ignoreCount;
        return this;
    }

    public BreakpointInfo setIgnoreCount(String ignoreCountStr) {
        try {
            this.ignoreCount = Integer.parseInt(ignoreCountStr);
        } catch (NumberFormatException ex) {
            this.ignoreCount = 0;
        }
        return this;
    }

    public String getLogMessage() {
        return logMessage;
    }

    public BreakpointInfo setLogMessage(String logMessage) {
        this.logMessage = logMessage;
        return this;
    }

    public boolean isForeign() {
        return foreign;
    }

    public BreakpointInfo setForeign(boolean foreign) {
        this.foreign = foreign;
        return this;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        BreakpointInfo that = (BreakpointInfo) o;
        if (foreign) {
            return that.foreign && Objects.equals(bpIdString, that.bpIdString);
        }
        return line == that.line
                && column == that.column
                && ignoreCount == that.ignoreCount
                && type == that.type
                && Objects.equals(file, that.file)
                && Objects.equals(function, that.function)
                && Objects.equals(condition, that.condition)
                && Objects.equals(logMessage, that.logMessage);
    }

    @Override
    public int hashCode() {
        return Objects.hash(type, enabled, file, line, column, function, condition, ignoreCount, logMessage);
    }
}
