/*******************************************************************************
 * Copyright (c) 2008-2019 Wind River Systems, Inc. and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License 2.0
 * which accompanies this distribution, and is available at
 * https://www.eclipse.org/legal/epl-2.0/
 *
 * Contributors:
 *     Wind River Systems - initial API and implementation
 *******************************************************************************/

/*******************************************************************************
 * Modified by Xilinx,Inc
 * Copyright 2022 Xilinx, Inc. All Rights Reserved. Modifications
 * are made available under the terms of the Eclipse Public License 2.0
 * which accompanies this distribution, and is available at
 * https://www.eclipse.org/legal/epl-2.0/
 *
 *******************************************************************************/

package org.eclipse.tcf.internal.dap.breakpoint;

import org.apache.commons.lang3.StringUtils;
import org.eclipse.lsp4j.debug.*;
import org.eclipse.lsp4j.debug.services.IDebugProtocolClient;
import org.eclipse.tcf.internal.dap.*;
import org.eclipse.tcf.internal.dap.actions.TCFAction;
import org.eclipse.tcf.internal.dap.breakpoint.BreakpointInfo.Type;
import org.eclipse.tcf.protocol.IChannel;
import org.eclipse.tcf.protocol.JSON;
import org.eclipse.tcf.protocol.Protocol;
import org.eclipse.tcf.services.IBreakpoints;
import org.eclipse.tcf.services.ILineNumbers;
import org.eclipse.tcf.services.IRunControl;
import org.eclipse.tcf.util.TCFDataCache;

import java.math.BigInteger;
import java.nio.file.Paths;
import java.util.*;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.function.Predicate;
import java.util.logging.Logger;
import java.util.stream.Collectors;

public class TCFBreakpointHandler {

    protected static final Logger LOGGER = Logger.getLogger(TCFDebugAdapter.class.getSimpleName());
    private final IChannel channel;
    private int nextBreakpointId = 0;
    private List<BreakpointInfo> bpInfoCache = new ArrayList<>();

    public TCFBreakpointHandler(IChannel channel) {
        this.channel = channel;
    }

    public BreakpointInfo[] setBreakpoints(final SetBreakpointsArguments args) {
        LOGGER.info(">> setBreakpoints");
        List<BreakpointInfo> bpInfo = getSourceBreakpointsInfo(args);
        String normalizedSourcePath = BreakpointUtils.getNormalizedSourcePath(args.getSource());
        CompletableFuture<BreakpointInfo[]> future = new CompletableFuture<>();
        TCFBreakpointsModel.INSTANCE.setBreakpoints(bpInfo, getSourceBreakpointPredicate(normalizedSourcePath), channel, future);
        try {
            BreakpointInfo[] actualBpInfo = future.get();
            return actualBpInfo;
        } catch (InterruptedException | ExecutionException ex) {
            ex.printStackTrace();
            return null;
        }
    }

    public BreakpointInfo[] setFunctionBreakpoints(final SetFunctionBreakpointsArguments args) {
        LOGGER.info(">> setFunctionBreakpoints");
        List<BreakpointInfo> bpInfo = getFunctionBreakpointsInfo(args);
        CompletableFuture<BreakpointInfo[]> future = new CompletableFuture<>();
        TCFBreakpointsModel.INSTANCE.setBreakpoints(bpInfo, getFunctionBreakpointPredicate(), channel, future);
        try {
            BreakpointInfo[] actualBpInfo = future.get();
            return actualBpInfo;
        } catch (InterruptedException | ExecutionException ex) {
            ex.printStackTrace();
            return null;
        }
    }

    public Breakpoint createBreakpointFromInfo(BreakpointInfo info) {
        Breakpoint[] breakpoint = createBreakpointsFromInfo(new BreakpointInfo[]{info});
        if (breakpoint == null || breakpoint.length == 0) {
            return null;
        }
        return breakpoint[0];
    }

    public Breakpoint[] createBreakpointsFromInfo(BreakpointInfo[] info) {
        List<Breakpoint> breakpoints = new ArrayList<>();
        for (BreakpointInfo bpInfo : info) {
            Breakpoint breakpoint = new Breakpoint();
            breakpoint.setId(bpInfo.getId());
            breakpoint.setVerified(!bpInfo.isForeign());
            if (bpInfo.isForeign()) {
                breakpoint.setMessage("Breakpoint set from another client");
            }
/*            IResource resource = ResourcesPlugin.getWorkspace().getRoot();
            if (markerAttrs.get(ICWatchpoint.EXPRESSION) != null) {
                bp = new CWatchpoint(resource, markerAttrs, true);
            }
            else if (markerAttrs.get(ICLineBreakpoint.ADDRESS) != null) {
                bp = new CAddressBreakpoint(resource, markerAttrs, true);
            }
            else*/
            if (StringUtils.isNotBlank(bpInfo.getFunction())) {
                Source source = new Source();
                source.setName(bpInfo.getFunction());
                breakpoint.setSource(source);
            }
            else if (StringUtils.isNotBlank(bpInfo.getFile())) {
                Source source = new Source();
                if (!bpInfo.isForeign()) {
                    source.setPath(bpInfo.getFile());
                    source.setName(Paths.get(bpInfo.getFile()).getFileName().toString());
                } else {
                    source.setName(bpInfo.getFile());
                }
                breakpoint.setSource(source);
                Integer line = bpInfo.getLine();
                if (line != null) {
                    breakpoint.setLine(line);
                    Integer column = bpInfo.getColumn();
                    if (column != null) {
                        breakpoint.setColumn(column);
                        breakpoint.setEndColumn(column + 1);
                    }
                }
            }
            breakpoints.add(breakpoint);
        }
        return breakpoints.toArray(new Breakpoint[0]);
    }

    public Predicate<BreakpointInfo> getSourceBreakpointPredicate(String filePath) {
        return (bp) -> bp.getType() == Type.SOURCE && bp.getFile().equals(filePath);
    }

    public Predicate<BreakpointInfo> getFunctionBreakpointPredicate() {
        return (bp) -> bp.getType() == Type.FUNCTION;
    }

    public List<BreakpointInfo> getSourceBreakpointsInfo(SetBreakpointsArguments args) {
        String srcPath = BreakpointUtils.getNormalizedSourcePath(args.getSource());
        List<BreakpointInfo> bpInfo = Arrays.stream(args.getBreakpoints())
                .map(bp -> getSourceBreakpointInfo(srcPath, bp))
                .collect(Collectors.toList());
        return bpInfo;
    }

    private BreakpointInfo getSourceBreakpointInfo(String file, SourceBreakpoint bp) {
        Optional<BreakpointInfo> cachedBreakpointInfo = bpInfoCache.stream()
                .filter(bpInfo -> bpInfo.getType() == Type.SOURCE
                                && Objects.equals(bpInfo.getFile(), file)
                                && Objects.equals(bpInfo.getLine(), bp.getLine())
                                && Objects.equals(bpInfo.getColumn(), bp.getColumn()))
                .findFirst();
        if (cachedBreakpointInfo.isPresent()) {
            return cachedBreakpointInfo.get()
                    .setCondition(bp.getCondition())
                    .setIgnoreCount(bp.getHitCondition())
                    .setLogMessage(bp.getLogMessage());
        }
        else {
            BreakpointInfo breakpointInfo = createSourceBreakpointInfo(file, bp);
            addBpInfoToCache(breakpointInfo);
            return breakpointInfo;
        }
    }

    private BreakpointInfo createSourceBreakpointInfo(String file, SourceBreakpoint bp) {
        BreakpointInfo breakpointInfo = BreakpointInfo.newLocalBreakpoint(Type.SOURCE, nextBreakpointId())
                .setFile(file)
                .setLine(bp.getLine())
                .setColumn(bp.getColumn())
                .setCondition(bp.getCondition())
                .setIgnoreCount(bp.getHitCondition())
                .setLogMessage(bp.getLogMessage());
        return breakpointInfo;
    }

    public List<BreakpointInfo> getFunctionBreakpointsInfo(SetFunctionBreakpointsArguments args) {
        List<BreakpointInfo> bpInfo = Arrays.stream(args.getBreakpoints())
                .map(bp -> getFunctionBreakpointInfo(bp))
                .collect(Collectors.toList());
        return bpInfo;
    }

    private BreakpointInfo getFunctionBreakpointInfo(FunctionBreakpoint bp) {
        Optional<BreakpointInfo> cachedBreakpointInfo = bpInfoCache.stream()
                .filter(bpInfo ->
                        bpInfo.getType() == Type.FUNCTION
                        && bpInfo.getFunction().equals(bp.getName()))
                .findFirst();
        if (cachedBreakpointInfo.isPresent()) {
            return cachedBreakpointInfo.get()
                    .setCondition(bp.getCondition())
                    .setIgnoreCount(bp.getHitCondition());
        }
        else {
            BreakpointInfo breakpointInfo = createFunctionBreakpointInfo(bp);
            addBpInfoToCache(breakpointInfo);
            return breakpointInfo;
        }
    }

    private BreakpointInfo createFunctionBreakpointInfo(FunctionBreakpoint bp) {
        BreakpointInfo breakpointInfo = BreakpointInfo.newLocalBreakpoint(Type.FUNCTION, nextBreakpointId())
                .setFunction(bp.getName())
                .setCondition(bp.getCondition())
                .setIgnoreCount(bp.getHitCondition());
        return breakpointInfo;
    }

    public void addBpInfoToCache(BreakpointInfo breakpointInfo) {
        BreakpointInfo copy = new BreakpointInfo();
        copy.setId(breakpointInfo.getId());
        copy.setForeign(breakpointInfo.isForeign());
        copy.setType(breakpointInfo.getType());
        copy.setFile(breakpointInfo.getFile());
        copy.setLine(breakpointInfo.getLine());
        copy.setColumn(breakpointInfo.getColumn());
        copy.setFunction(breakpointInfo.getFunction());
        bpInfoCache.add(copy);
    }

    public synchronized int nextBreakpointId() {
        return nextBreakpointId++;
    }

    private Map<String, Integer> adjustedBreakpointLines = new HashMap<>();

    public void fireBreakpointLineUpdates(BreakpointInfo[] bpInfo, List<TCFNodeExecContext> threads,
                                          TCFBreakpointsStatus bpStatus, IDebugProtocolClient client,
                                          boolean isNewBreakpointEvent) {
        Protocol.invokeLater(200, new Runnable() {
            @Override
            public void run() {
                Map<String, Integer> idLineMap = new HashMap<>();
                for (BreakpointInfo originalInfo : bpInfo) {
                    String id = originalInfo.getBreakpointId();
                    Integer adjustedStartLine = adjustedBreakpointLines.get(id);
                    if (adjustedStartLine == null) {
                        Map<String, Object> status = bpStatus.getStatus(id);
                        if (status == null) continue;
                        String error = (String) status.get(IBreakpoints.STATUS_ERROR);
                        if (error != null) continue;
                        Object[] arr = toObjectArray(status.get(IBreakpoints.STATUS_INSTANCES));
                        if (arr == null) continue;
                        for (Object o : arr) {
                            Map<String, Object> m = toObjectMap(o);
                            String ctx_id = (String) m.get(IBreakpoints.INSTANCE_CONTEXT);
                            if (ctx_id == null) continue;

                            TCFNodeExecContext thread = null;
                            TCFNodeExecContext memory = null;
                            for (TCFNodeExecContext node : threads) {
                                TCFNodeExecContext nodeMemory = null;
                                String bpGroup = null;
                                TCFAction action = node.getModel().getActiveAction(node.getID());
                                if (action == null || !action.showRunning()) {
                                    TCFDataCache<IRunControl.RunControlContext> rc_ctx_cache = node.getRunContext();
                                    if (!rc_ctx_cache.validate(this)) return;
                                    IRunControl.RunControlContext rc_ctx_data = rc_ctx_cache.getData();
                                    if (rc_ctx_data != null) bpGroup = rc_ctx_data.getBPGroup();
                                    TCFDataCache<TCFNodeExecContext> mem_cache = node.getMemoryNode();
                                    if (!mem_cache.validate(this)) return;
                                    nodeMemory = mem_cache.getData();
                                    if (bpGroup == null && nodeMemory != null && rc_ctx_data != null && rc_ctx_data.hasState())
                                        bpGroup = nodeMemory.getID();
                                    if (ctx_id.equals(node.getID()) || ctx_id.equals(bpGroup)) {
                                        thread = node;
                                        memory = nodeMemory;
                                        break;
                                    }
                                }
                            }

                            if (thread != null && memory != null) {
                                TCFLaunch launch = thread.getModel().getLaunch();
                                BigInteger addr = JSON.toBigInteger((Number) m.get(IBreakpoints.INSTANCE_ADDRESS));
                                ILineNumbers.CodeArea area = null;
                                ILineNumbers.CodeArea org_area = getBreakpointCodeArea(launch, id);
                                if (addr != null) {
                                    TCFDataCache<TCFSourceRef> line_cache = memory.getLineInfo(addr);
                                    if (line_cache != null) {
                                        if (!line_cache.validate(this)) return;
                                        TCFSourceRef line_data = line_cache.getData();
                                        if (line_data != null) area = line_data.area;
                                    }
                                }
                                if (area == null) area = org_area;
//								String bp_name = "Breakpoint";
//								if (bp != null) bp_name = bp.getMarker().getAttribute(TCFBreakpointsModel.ATTR_MESSAGE, bp_name);
                                if (area != null && addr != null) {
                                    if (isLineAdjusted(area, org_area)) {
                                        adjustedStartLine = area.start_line;
                                        adjustedBreakpointLines.put(id, adjustedStartLine);
                                    }
                                }
                            }
                        }
                    }
                    if (adjustedStartLine != null) {
                        idLineMap.put(id, adjustedStartLine);
                    }
                }
                for (BreakpointInfo info : bpInfo) {
                    Integer newStartLine = adjustedBreakpointLines.get(info.getBreakpointId());
                    if (newStartLine == null) {
                        continue;
                    }
                    Breakpoint breakpoint = createBreakpointFromInfo(info);
                    if (breakpoint != null) {
                        breakpoint.setLine(newStartLine);
                        BreakpointEventArguments eventArgs = new BreakpointEventArguments();
                        eventArgs.setBreakpoint(breakpoint);
                        eventArgs.setReason(isNewBreakpointEvent ? BreakpointEventArgumentsReason.NEW : BreakpointEventArgumentsReason.CHANGED);
                        client.breakpoint(eventArgs);
                    }
                }
            }
        });
    }

    private boolean isLineAdjusted(ILineNumbers.CodeArea x, ILineNumbers.CodeArea y) {
        if (x == null) return y != null;
        if (y == null) return x != null;
        if (x.start_line == 0 || y.start_line == 0) return false;
        if (x.start_line != y.start_line) return true;
        return false;
    }

    @SuppressWarnings("unchecked")
    private Object[] toObjectArray(Object o) {
        if (o == null) return null;
        Collection<Object> c = (Collection<Object>)o;
        return (Object[])c.toArray(new Object[c.size()]);
    }

    @SuppressWarnings("unchecked")
    private Map<String,Object> toObjectMap(Object o) {
        return (Map<String,Object>)o;
    }

    private ILineNumbers.CodeArea getBreakpointCodeArea(TCFLaunch launch, String id) {
        Map<String,Object> props = launch.getBreakpointsStatus().getProperties(id);
        if (props != null && props.get(IBreakpoints.PROP_LOCATION) == null) {
            String file = (String)props.get(IBreakpoints.PROP_FILE);
            Number line = (Number)props.get(IBreakpoints.PROP_LINE);
            if (file != null && line != null) {
                return new ILineNumbers.CodeArea(null, file,
                        line.intValue(), 0, line.intValue() + 1, 0,
                        null, null, 0, false, false, false, false);
            }
        }
        return null;
    }
}
